#!/usr/bin/env python3

"""
test.unit_tests_d.ut_msg: message signing unit tests for the MMGen suite
"""

import os

from mmgen.util import msg,bmsg,pumsg,suf
from mmgen.protocol import CoinProtocol
from mmgen.msg import NewMsg,UnsignedMsg,SignedMsg,SignedOnlineMsg,ExportedMsgSigs
from mmgen.addr import MMGenID

from ..include.common import cfg,silence,end_silence,restart_test_daemons,stop_test_daemons

def get_obj(coin,network,msghash_type):

	if coin == 'bch':
		addrlists = 'DEADBEEF:C:1-20 98831F3A:C:8,2 A091ABAA:L:111 A091ABAA:C:1'
	elif coin == 'eth':
		addrlists = 'DEADBEEF:E:1-20 98831F3A:E:8,2 A091ABAA:E:111'
	else:
		# A091ABAA = 98831F3A:5S
		addrlists = 'DEADBEEF:C:1-20 98831F3A:B:8,2 A091ABAA:S:10-11 A091ABAA:111 A091ABAA:C:1'

	return NewMsg(
		cfg       = cfg,
		coin      = coin,
		network   = network,
		message   = '08/Jun/2021 Bitcoin Law Enacted by El Salvador Legislative Assembly',
		addrlists = addrlists,
		msghash_type = msghash_type )

def print_total(n):
	msg(f'{n} signature{suf(n)} verified')

async def run_test(network_id,chksum,msghash_type='raw'):

	coin,network = CoinProtocol.Base.parse_network_id(network_id)

	if not cfg.verbose:
		silence()

	bmsg(f'\nTesting {coin.upper()} {network.upper()}:\n')

	m = get_obj(coin,network,msghash_type)

	if m.proto.sign_mode == 'daemon':
		restart_test_daemons(network_id)

	pumsg('\nTesting data creation:\n')

	tmpdir = os.path.join('test','trash2')

	os.makedirs(tmpdir,exist_ok=True)

	assert m.chksum.upper() == chksum, f'{m.chksum.upper()} != {chksum}'

	m.write_to_file(
		outdir        = tmpdir,
		ask_overwrite = False )

	pumsg('\nTesting signing:\n')

	m = UnsignedMsg( cfg, infile = os.path.join(tmpdir,get_obj(coin,network,msghash_type).filename) )
	await m.sign(wallet_files=['test/ref/98831F3A.mmwords'])

	m = SignedMsg( cfg, data=m.__dict__ )
	m.write_to_file(
		outdir        = tmpdir,
		ask_overwrite = False )

	pumsg('\nTesting display:\n')

	m = SignedOnlineMsg( cfg, infile = os.path.join(tmpdir,get_obj(coin,network,msghash_type).signed_filename) )

	msg(m.format())

	single_addr = 'A091ABAA:E:111' if m.proto.base_proto == 'Ethereum' else 'A091ABAA:111'
	single_addr_coin = m.sigs[MMGenID(m.proto,single_addr)]['addr']

	pumsg('\nTesting single address display:\n')
	msg(m.format(single_addr))

	pumsg('\nTesting verification:\n')
	print_total( await m.verify() )

	pumsg('\nTesting single address verification:\n')
	print_total( await m.verify(single_addr) )

	pumsg('\nTesting JSON dump for export:\n')
	msg( m.get_json_for_export() )

	pumsg('\nTesting single address JSON dump for export:\n')
	msg( m.get_json_for_export(single_addr) )

	from mmgen.fileutil import write_data_to_file
	exported_sigs = os.path.join(tmpdir,'signatures.json')
	write_data_to_file(
		cfg     = cfg,
		outfile = exported_sigs,
		data    = m.get_json_for_export(),
		desc    = 'signature data',
		ask_overwrite = False )

	m = ExportedMsgSigs( cfg, infile=exported_sigs )

	pumsg('\nTesting verification (exported data):\n')
	print_total( await m.verify() )

	pumsg('\nTesting single address verification (exported data):\n')
	print_total( await m.verify(single_addr_coin) )

	pumsg('\nTesting display (exported data):\n')
	msg(m.format())

	pumsg('\nTesting single address display (exported data):\n')
	msg(m.format(single_addr_coin))

	if m.proto.sign_mode == 'daemon':
		stop_test_daemons(network_id)

	msg('\n')

	if not cfg.verbose:
		end_silence()

	return True

class unit_tests:

	altcoin_deps = ('ltc','bch','eth','eth_raw')

	def btc(self,name,ut):
		return run_test('btc','AA0DB5')

	def btc_tn(self,name,ut):
		return run_test('btc_tn','A88E1D')

	def btc_rt(self,name,ut):
		return run_test('btc_rt','578018')

	def ltc(self,name,ut):
		return run_test('ltc','BA7549')

	def bch(self,name,ut):
		return run_test('bch','1B8065')

	def eth(self,name,ut):
		return run_test('eth','35BAD9',msghash_type='eth_sign')

	def eth_raw(self,name,ut):
		return run_test('eth','9D900C')



#!/usr/bin/env python3

"""
test.unit_tests_d.ut_lockable: unit test for the MMGen suite's Lockable class
"""

from ..include.common import qmsg,qmsg_r,vmsg

class unit_test:

	def run_test(self,name,ut):

		from mmgen.base_obj import AttrCtrl,Lockable
		from decimal import Decimal

		qmsg_r('Testing class AttrCtrl...')

		class MyAttrCtrl(AttrCtrl):
			_autolock = False
			foo = 'fooval'
		ac = MyAttrCtrl()
		ac._lock()

		ac.foo = 'new fooval'
		ac.foo = 'new fooval2'

		class MyAttrCtrlAutolock(AttrCtrl):
			pass

		aca = MyAttrCtrlAutolock()

		class MyAttrCtrlClsCheck(AttrCtrl):
			_autolock = False
			_use_class_attr = True
			foo = 'fooval'
			bar = None
		acc = MyAttrCtrlClsCheck()
		acc._lock()

		acc.foo = 'new_fooval'
		acc.foo = 'new_fooval2'
		acc.bar = 'bar val'
		acc.bar = 1 # class attribute bar is None, so can be set to any type

		class MyAttrCtrlDflNone(AttrCtrl):
			_default_to_none = True
			foo = 'fooval'
			bar = None

		acdn = MyAttrCtrlDflNone()
		assert acdn.foo == 'fooval', f'{acdn.foo}'
		assert acdn.bar is None, f'{acdn.bar}'
		assert acdn.baz is None, f'{acdn.baz}'

		qmsg('OK')
		qmsg_r('Testing class Lockable...')

		class MyLockable(Lockable): # class without attrs
			_autolock = False
			_set_ok = ('foo','baz','alpha','beta','gamma','delta','epsilon')
			_reset_ok = ('bar','baz')

		lc = MyLockable()
		lc.foo = None
		lc.bar = 'barval'
		lc.baz = 1
		lc.qux = 1

		# are these considered set?
		lc.alpha = 0             # yes
		lc.beta = False          # yes
		lc.gamma = Decimal('0')  # yes
		lc.delta = 0.0           # yes
		lc.epsilon = []          # no

		lc._lock()

		lc.foo = 'fooval2'
		lc.bar = 'barval2'
		lc.bar = 'barval3'
		lc.baz = 2
		lc.baz = 3

		lc.epsilon = [0]

		class MyLockableClsCheck(Lockable): # class with attrs
			_autolock = False
			_use_class_attr = True
			_set_ok = ('foo','baz')
			_reset_ok = ('bar','baz')
			foo = None
			bar = 1
			baz = 3.5
			qux = 'quxval'

		lcc = MyLockableClsCheck()
		lcc._lock()

		lcc.foo = 'fooval2' # class attribute foo is None, so can be set to any type
		lcc.bar = 2
		lcc.bar = 3   # bar is in reset list
		lcc.baz = 3.2
		lcc.baz = 3.1 # baz is in both lists

		qmsg('OK')
		qmsg_r('Testing class Lockable with autolock...')

		class MyLockableAutolock(Lockable):
			def __init__(self):
				self.foo = True

		lca = MyLockableAutolock()
		assert lca._autolock is True
		assert lca._locked is True
		assert lca.foo is True

		class MyLockableAutolockDflNone(Lockable):
			_default_to_none = True
			foo = 0

		lcdn = MyLockableAutolockDflNone()
		assert lcdn.foo == 0
		assert lcdn.bar == None

		class MyLockableBad(Lockable):
			_set_ok = ('foo','bar')
			foo = 1

		qmsg('OK')
		qmsg_r('Checking error handling...')
		vmsg('')

		def bad1(): ac.x = 1
		def bad2(): acc.foo = 1
		def bad3(): lc.foo = 'fooval3'
		def bad4(): lc.baz = 'str'
		def bad5(): lcc.bar = 'str'
		def bad6(): lc.qux  = 2
		def bad7(): lcc.qux = 'quxval2'
		def bad8(): lcc.foo = 'fooval3'
		def bad9(): lc.x = 1
		def bad10(): lcc.x = 1

		def bad11(): lc.alpha = 0
		def bad12(): lc.beta = False
		def bad13(): lc.gamma = Decimal('0')
		def bad14(): lc.delta = float(0)
		def bad15(): lc.epsilon = [0]

		def bad16(): lca.foo = None
		def bad17(): lb = MyLockableBad()
		def bad18(): aca._lock()

		def bad19(): acdn.baz = None
		def bad20(): lcdn.foo = 1
		def bad21(): lcdn.bar = None
		def bad22(): del lcdn.foo

		ut.process_bad_data((
			('attr (1)',           'AttributeError', 'has no attr', bad1 ),
			('attr (2)',           'AttributeError', 'has no attr', bad9 ),
			('attr (3)',           'AttributeError', 'has no attr', bad10 ),
			('attr (4)',           'AttributeError', 'has no attr', bad19 ),
			('attr (5)',           'AttributeError', 'has no attr', bad21 ),
			('attr type (1)',      'AttributeError', 'type',        bad2 ),
			("attr type (2)",      'AttributeError', 'type',        bad4 ),
			("attr type (3)",      'AttributeError', 'type',        bad5 ),
			("attr (can't set)",   'AttributeError', 'read-only',   bad6 ),
			("attr (can't set)",   'AttributeError', 'read-only',   bad7 ),
			("attr (can't set)",   'AttributeError', 'read-only',   bad20 ),
			("attr (can't reset)", 'AttributeError', 'reset',       bad3 ),
			("attr (can't reset)", 'AttributeError', 'reset',       bad8 ),
			("attr (can't reset)", 'AttributeError', 'reset',       bad11 ),
			("attr (can't reset)", 'AttributeError', 'reset',       bad12 ),
			("attr (can't reset)", 'AttributeError', 'reset',       bad13 ),
			("attr (can't reset)", 'AttributeError', 'reset',       bad14 ),
			("attr (can't reset)", 'AttributeError', 'reset',       bad15 ),
			("attr (can't set)",   'AttributeError', 'read-only',   bad16 ),
			("attr (bad _set_ok)", 'AssertionError', 'not found in',bad17 ),
			("attr (can’t delete)",'AttributeError', 'not be delet',bad22 ),
			("call to _lock()",    'AssertionError', 'only once',   bad18 ),
		))

		qmsg('OK')
		return True



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tx.sign: Sign a transaction generated by 'mmgen-txcreate'
"""

from ..cfg import gc
from ..util import msg,suf,fmt,die,remove_dups,get_extension
from ..obj import MMGenList
from ..addr import MMGenAddrType
from ..addrlist import AddrIdxList,KeyAddrList
from ..wallet import Wallet,get_wallet_extensions,get_wallet_cls

saved_seeds = {}

def get_seed_for_seed_id(sid,infiles,saved_seeds):

	if sid in saved_seeds:
		return saved_seeds[sid]

	subseeds_checked = False
	while True:
		if infiles:
			seed = Wallet(cfg, infiles.pop(0), ignore_in_fmt=True, passwd_file=global_passwd_file).seed
		elif subseeds_checked is False:
			seed = saved_seeds[list(saved_seeds)[0]].subseed_by_seed_id(sid,print_msg=True)
			subseeds_checked = True
			if not seed:
				continue
		elif cfg.in_fmt:
			cfg._util.qmsg(f'Need seed data for Seed ID {sid}')
			seed = Wallet(cfg, passwd_file=global_passwd_file).seed
			msg(f'User input produced Seed ID {seed.sid}')
			if not seed.sid == sid: # TODO: add test
				seed = seed.subseed_by_seed_id(sid,print_msg=True)

		if seed:
			saved_seeds[seed.sid] = seed
			if seed.sid == sid:
				return seed
		else:
			die(2,f'ERROR: No seed source found for Seed ID: {sid}')

def generate_kals_for_mmgen_addrs(need_keys,infiles,saved_seeds,proto):
	mmids = [e.mmid for e in need_keys]
	sids = remove_dups((i.sid for i in mmids),quiet=True)
	cfg._util.vmsg(f"Need seed{suf(sids)}: {' '.join(sids)}")
	def gen_kals():
		for sid in sids:
			# Returns only if seed is found
			seed = get_seed_for_seed_id(sid,infiles,saved_seeds)
			for id_str in MMGenAddrType.mmtypes:
				idx_list = [i.idx for i in mmids if i.sid == sid and i.mmtype == id_str]
				if idx_list:
					yield KeyAddrList(
						cfg       = cfg,
						proto     = proto,
						seed      = seed,
						addr_idxs = AddrIdxList(idx_list=idx_list),
						mmtype    = MMGenAddrType(proto,id_str),
						skip_chksum = True )
	return MMGenList(gen_kals())

def add_keys(tx,src,infiles=None,saved_seeds=None,keyaddr_list=None):
	need_keys = [e for e in getattr(tx,src) if e.mmid and not e.have_wif]
	if not need_keys:
		return []
	desc,src_desc = (
		('key-address file','From key-address file:') if keyaddr_list else
		('seed(s)','Generated from seed:') )
	cfg._util.qmsg(f'Checking {gc.proj_name} -> {tx.proto.coin} address mappings for {src} (from {desc})')
	d = (
		MMGenList([keyaddr_list]) if keyaddr_list else
		generate_kals_for_mmgen_addrs(need_keys,infiles,saved_seeds,tx.proto) )
	new_keys = []
	for e in need_keys:
		for kal in d:
			for f in kal.data:
				mmid = f'{kal.al_id}:{f.idx}'
				if mmid == e.mmid:
					if f.addr == e.addr:
						e.have_wif = True
						if src == 'inputs':
							new_keys.append(f)
					else:
						die(3,fmt(f"""
							{gc.proj_name} -> {tx.proto.coin} address mappings differ!
							{src_desc:<23} {mmid} -> {f.addr}
							{'tx file:':<23} {e.mmid} -> {e.addr}
							""").strip())
	if new_keys:
		cfg._util.vmsg(f'Added {len(new_keys)} wif key{suf(new_keys)} from {desc}')
	return new_keys

def _pop_matching_fns(args,cmplist): # strips found args
	return list(reversed([args.pop(args.index(a)) for a in reversed(args) if get_extension(a) in cmplist]))

def get_tx_files(cfg, args):
	from .unsigned import Unsigned, AutomountUnsigned
	ret = _pop_matching_fns(args, [(AutomountUnsigned if cfg.autosign else Unsigned).ext])
	if not ret:
		die(1,'You must specify a raw transaction file!')
	return ret

def get_seed_files(cfg,args):
	# favor unencrypted seed sources first, as they don't require passwords
	ret = _pop_matching_fns( args, get_wallet_extensions('unenc') )
	from ..filename import find_file_in_dir
	wf = find_file_in_dir(get_wallet_cls('mmgen'),cfg.data_dir) # Make this the first encrypted ss in the list
	if wf:
		ret.append(wf)
	ret += _pop_matching_fns( args, get_wallet_extensions('enc') )
	if not (ret or cfg.mmgen_keys_from_file or cfg.keys_from_file): # or cfg.use_wallet_dat
		die(1,'You must specify a seed or key source!')
	return ret

def get_keyaddrlist(cfg,proto):
	if cfg.mmgen_keys_from_file:
		return KeyAddrList( cfg, proto, cfg.mmgen_keys_from_file )
	return None

def get_keylist(cfg):
	if cfg.keys_from_file:
		from ..fileutil import get_lines_from_file
		return get_lines_from_file( cfg, cfg.keys_from_file, 'key-address data', trim_comments=True )
	return None

async def txsign(cfg_parm, tx, seed_files, kl, kal, tx_num_str='', passwd_file=None):

	keys = MMGenList() # list of AddrListEntry objects
	non_mmaddrs = tx.get_non_mmaddrs('inputs')

	global cfg, global_passwd_file
	cfg = cfg_parm
	global_passwd_file = passwd_file

	if non_mmaddrs:
		tx.check_non_mmgen_inputs(caller='txsign',non_mmaddrs=non_mmaddrs)
		tmp = KeyAddrList(
			cfg         = cfg,
			proto       = tx.proto,
			addrlist    = non_mmaddrs,
			skip_chksum = True )
		if kl:
			tmp.add_wifs(kl)
		missing = tmp.list_missing('sec')
		if missing:
			sep = '\n    '
			die(2,'ERROR: a key file must be supplied for the following non-{} address{}:{}'.format(
				gc.proj_name,
				suf(missing,'es'),
				sep + sep.join(missing) ))
		keys += tmp.data

	if cfg.mmgen_keys_from_file:
		keys += add_keys(tx,'inputs',keyaddr_list=kal)
		add_keys(tx,'outputs',keyaddr_list=kal)

	keys += add_keys(tx,'inputs',seed_files,saved_seeds)
	add_keys(tx,'outputs',seed_files,saved_seeds)

	# this (boolean) attr isn't needed in transaction file
	tx.delete_attrs('inputs','have_wif')
	tx.delete_attrs('outputs','have_wif')

	extra_sids = remove_dups(
		(s for s in saved_seeds if s not in tx.get_sids('inputs') + tx.get_sids('outputs')),
		quiet = True )

	if extra_sids:
		msg(f"Unused Seed ID{suf(extra_sids)}: {' '.join(extra_sids)}")

	return await tx.sign(tx_num_str,keys) # returns signed TX object or False



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.incog: incognito wallet class
"""

from .incog_base import wallet

class wallet(wallet):

	desc = 'incognito data'
	file_mode = 'binary'
	no_tty = True



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tx.file: Transaction file operations for the MMGen suite
"""
import os

from ..util import ymsg,make_chksum_6,die
from ..obj import MMGenObject,HexStr,MMGenTxID,CoinTxID,MMGenTxComment

class MMGenTxFile(MMGenObject):

	def __init__(self,tx):
		self.tx       = tx
		self.chksum   = None
		self.fmt_data = None
		self.filename = None

	def parse(self,infile,metadata_only=False,quiet_open=False):
		tx = self.tx

		def eval_io_data(raw_data,desc):
			from ast import literal_eval
			try:
				d = literal_eval(raw_data)
			except:
				if desc == 'inputs' and not quiet_open:
					ymsg('Warning: transaction data appears to be in old format')
				import re
				d = literal_eval(re.sub(r"[A-Za-z]+?\(('.+?')\)",r'\1',raw_data))
			assert isinstance(d,list), f'{desc} data not a list!'
			if not (desc == 'outputs' and tx.proto.base_coin == 'ETH'): # ETH txs can have no outputs
				assert len(d), f'no {desc}!'
			for e in d:
				e['amt'] = tx.proto.coin_amt(e['amt'])
				if 'label' in e:
					e['comment'] = e['label']
					del e['label']
			io,io_list = {
				'inputs':  ( tx.Input, tx.InputList ),
				'outputs': ( tx.Output, tx.OutputList ),
			}[desc]
			return io_list( parent=tx, data=[io(tx.proto,**e) for e in d] )

		from ..fileutil import get_data_from_file
		tx_data = get_data_from_file( tx.cfg, infile, tx.desc+' data', quiet=quiet_open )

		desc = 'data'
		try:
			if len(tx_data) > tx.cfg.max_tx_file_size:
				die('MaxFileSizeExceeded',
					f'Transaction file size exceeds limit ({tx.cfg.max_tx_file_size} bytes)')
			tx_data = tx_data.splitlines()
			assert len(tx_data) >= 5,'number of lines less than 5'
			assert len(tx_data[0]) == 6,'invalid length of first line'
			self.chksum = HexStr(tx_data.pop(0))
			assert self.chksum == make_chksum_6(' '.join(tx_data)),'file data does not match checksum'

			if len(tx_data) == 7:
				desc = 'sent timestamp'
				(_, tx.sent_timestamp) = tx_data.pop(-1).split()
				assert _ == 'Sent', 'invalid sent timestamp line'

			if len(tx_data) == 6:
				assert len(tx_data[-1]) == 64,'invalid coin TxID length'
				desc = 'coin TxID'
				tx.coin_txid = CoinTxID(tx_data.pop(-1))

			if len(tx_data) == 5:
				# rough check: allow for 4-byte utf8 characters + base58 (4 * 11 / 8 = 6 (rounded up))
				assert len(tx_data[-1]) < MMGenTxComment.max_len*6,'invalid comment length'
				c = tx_data.pop(-1)
				if c != '-':
					desc = 'encoded comment (not base58)'
					from ..baseconv import baseconv
					comment = baseconv('b58').tobytes(c).decode()
					assert comment is not False,'invalid comment'
					desc = 'comment'
					tx.comment = MMGenTxComment(comment)

			desc = 'number of lines' # four required lines
			( metadata, tx.serialized, inputs_data, outputs_data ) = tx_data
			assert len(metadata) < 100,'invalid metadata length' # rough check
			metadata = metadata.split()

			if metadata[-1].startswith('LT='):
				desc = 'locktime'
				tx.locktime = int(metadata.pop()[3:])

			desc = 'coin token in metadata'
			coin = metadata.pop(0) if len(metadata) == 6 else 'BTC'
			coin,tokensym = coin.split(':') if ':' in coin else (coin,None)

			desc = 'chain token in metadata'
			tx.chain = metadata.pop(0).lower() if len(metadata) == 5 else 'mainnet'

			from ..protocol import CoinProtocol,init_proto
			network = CoinProtocol.Base.chain_name_to_network(tx.cfg,coin,tx.chain)

			desc = 'initialization of protocol'
			tx.proto = init_proto( tx.cfg, coin, network=network, need_amt=True )
			if tokensym:
				tx.proto.tokensym = tokensym

			desc = 'metadata (4 items)'
			(txid, send_amt, tx.timestamp, blockcount) = metadata

			desc = 'TxID in metadata'
			tx.txid = MMGenTxID(txid)
			desc = 'block count in metadata'
			tx.blockcount = int(blockcount)

			if metadata_only:
				return

			desc = 'transaction file hex data'
			tx.check_txfile_hex_data()
			desc = 'Ethereum RLP or JSON data'
			tx.parse_txfile_serialized_data()
			desc = 'inputs data'
			tx.inputs  = eval_io_data(inputs_data,'inputs')
			desc = 'outputs data'
			tx.outputs = eval_io_data(outputs_data,'outputs')
			desc = 'send amount in metadata'
			from decimal import Decimal
			assert Decimal(send_amt) == tx.send_amt, f'{send_amt} != {tx.send_amt}'
		except Exception as e:
			die(2,f'Invalid {desc} in transaction file: {e!s}')

	def make_filename(self):
		tx = self.tx
		def gen_filename():
			yield tx.txid
			if tx.coin != 'BTC':
				yield '-' + tx.dcoin
			yield f'[{tx.send_amt!s}'
			if tx.is_replaceable():
				yield ',{}'.format(tx.fee_abs2rel(tx.fee,to_unit=tx.fn_fee_unit))
			if tx.get_serialized_locktime():
				yield f',tl={tx.get_serialized_locktime()}'
			yield ']'
			if tx.proto.testnet:
				yield '.' + tx.proto.network
			yield '.' + tx.ext
		return ''.join(gen_filename())

	def format(self):
		tx = self.tx

		def amt_to_str(d):
			return {k: (str(d[k]) if k == 'amt' else d[k]) for k in d}

		coin_id = '' if tx.coin == 'BTC' else tx.coin + ('' if tx.coin == tx.dcoin else ':'+tx.dcoin)
		lines = [
			'{}{} {} {} {} {}{}'.format(
				(coin_id+' ' if coin_id else ''),
				tx.chain.upper(),
				tx.txid,
				tx.send_amt,
				tx.timestamp,
				tx.blockcount,
				(f' LT={tx.locktime}' if tx.locktime else ''),
			),
			tx.serialized,
			ascii([amt_to_str(e._asdict()) for e in tx.inputs]),
			ascii([amt_to_str(e._asdict()) for e in tx.outputs])
		]

		if tx.comment:
			from ..baseconv import baseconv
			lines.append(baseconv('b58').frombytes(tx.comment.encode(),tostr=True))

		if tx.coin_txid:
			if not tx.comment:
				lines.append('-') # keep old tx files backwards compatible
			lines.append(tx.coin_txid)

		if tx.sent_timestamp:
			lines.append(f'Sent {tx.sent_timestamp}')

		self.chksum = make_chksum_6(' '.join(lines))
		fmt_data = '\n'.join([self.chksum] + lines) + '\n'
		if len(fmt_data) > tx.cfg.max_tx_file_size:
			die( 'MaxFileSizeExceeded', f'Transaction file size exceeds limit ({tx.cfg.max_tx_file_size} bytes)' )
		return fmt_data

	def write(self,
		add_desc              = '',
		outdir                = None,
		ask_write             = True,
		ask_write_default_yes = False,
		ask_tty               = True,
		ask_overwrite         = True ):

		if ask_write is False:
			ask_write_default_yes = True

		if not self.filename:
			self.filename = self.make_filename()

		if not self.fmt_data:
			self.fmt_data = self.format()

		from ..fileutil import write_data_to_file
		write_data_to_file(
			cfg                   = self.tx.cfg,
			outfile               = os.path.join((outdir or ''), self.filename),
			data                  = self.fmt_data,
			desc                  = self.tx.desc + add_desc,
			ask_overwrite         = ask_overwrite,
			ask_write             = ask_write,
			ask_tty               = ask_tty,
			ask_write_default_yes = ask_write_default_yes,
			ignore_opt_outdir     = outdir)

	@classmethod
	def get_proto(cls,cfg,filename,quiet_open=False):
		from . import BaseTX
		tmp_tx = BaseTX(cfg=cfg)
		cls(tmp_tx).parse(filename,metadata_only=True,quiet_open=quiet_open)
		return tmp_tx.proto



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
proto.eth.contract: Ethereum ERC20 token classes
"""

from decimal import Decimal
from . import rlp

from . import erigon_sleep
from ...util import msg,pp_msg,die
from ...base_obj import AsyncInit
from ...obj import MMGenObject,CoinTxID
from ...addr import CoinAddr,TokenAddr

def parse_abi(s):
	return [s[:8]] + [s[8+x*64:8+(x+1)*64] for x in range(len(s[8:])//64)]

class TokenCommon(MMGenObject):

	def create_method_id(self,sig):
		return self.keccak_256(sig.encode()).hexdigest()[:8]

	def transferdata2sendaddr(self,data): # online
		return CoinAddr(self.proto,parse_abi(data)[1][-40:])

	def transferdata2amt(self,data): # online
		return self.proto.coin_amt(int(parse_abi(data)[-1],16) * self.base_unit)

	async def do_call(self,method_sig,method_args='',toUnit=False):
		data = self.create_method_id(method_sig) + method_args
		if self.cfg.debug:
			msg('ETH_CALL {}:  {}'.format(
				method_sig,
				'\n  '.join(parse_abi(data)) ))
		ret = await self.rpc.call('eth_call',{ 'to': '0x'+self.addr, 'data': '0x'+data },'pending')
		await erigon_sleep(self)
		if toUnit:
			return int(ret,16) * self.base_unit
		else:
			return ret

	async def get_balance(self,acct_addr):
		return self.proto.coin_amt(await self.do_call('balanceOf(address)',acct_addr.rjust(64,'0'),toUnit=True))

	def strip(self,s):
		return ''.join([chr(b) for b in s if 32 <= b <= 127]).strip()

	async def get_name(self):
		return self.strip(bytes.fromhex((await self.do_call('name()'))[2:]))

	async def get_symbol(self):
		return self.strip(bytes.fromhex((await self.do_call('symbol()'))[2:]))

	async def get_decimals(self):
		ret = await self.do_call('decimals()')
		try:
			assert ret[:2] == '0x'
			return int(ret,16)
		except:
			msg(f'RPC call to decimals() failed (returned {ret!r})')
			return None

	async def get_total_supply(self):
		return await self.do_call('totalSupply()',toUnit=True)

	async def info(self):
		return ('{:15}{}\n' * 5).format(
			'token address:', self.addr,
			'token symbol:',  await self.get_symbol(),
			'token name:',    await self.get_name(),
			'decimals:',      self.decimals,
			'total supply:',  await self.get_total_supply() )

	async def code(self):
		return (await self.rpc.call('eth_getCode','0x'+self.addr))[2:]

	def create_data(
			self,
			to_addr,
			amt,
			method_sig = 'transfer(address,uint256)'):
		from_arg = ''
		to_arg = to_addr.rjust(64,'0')
		amt_arg = '{:064x}'.format( int(amt / self.base_unit) )
		return self.create_method_id(method_sig) + from_arg + to_arg + amt_arg

	def make_tx_in(
			self,
			to_addr,
			amt,
			start_gas,
			gasPrice,
			nonce,
			method_sig = 'transfer(address,uint256)'):
		data = self.create_data(
				to_addr,
				amt,
				method_sig = method_sig)
		return {
			'to':       bytes.fromhex(self.addr),
			'startgas': start_gas.toWei(),
			'gasprice': gasPrice.toWei(),
			'value':    0,
			'nonce':    nonce,
			'data':     bytes.fromhex(data)}

	async def txsign(self,tx_in,key,from_addr,chain_id=None):

		from .pyethereum.transactions import Transaction

		if chain_id is None:
			res = await self.rpc.call('eth_chainId')
			chain_id = None if res is None else int(res,16)

		tx = Transaction(**tx_in).sign(key,chain_id)

		if tx.sender.hex() != from_addr:
			die(3,f'Sender address {from_addr!r} does not match address of key {tx.sender.hex()!r}!')

		if self.cfg.debug:
			msg('TOKEN DATA:')
			pp_msg(tx.to_dict())
			msg('PARSED ABI DATA:\n  {}'.format(
				'\n  '.join(parse_abi(tx.data.hex())) ))

		return (
			rlp.encode(tx).hex(),
			CoinTxID(tx.hash.hex())
		)

# The following are used for token deployment only:

	async def txsend(self,txhex):
		return (await self.rpc.call('eth_sendRawTransaction','0x'+txhex)).replace('0x','',1)

	async def transfer(
			self,
			from_addr,
			to_addr,
			amt,
			key,
			start_gas,
			gasPrice,
			method_sig = 'transfer(address,uint256)'):
		tx_in = self.make_tx_in(
				to_addr,
				amt,
				start_gas,
				gasPrice,
				nonce = int(await self.rpc.call('eth_getTransactionCount','0x'+from_addr,'pending'),16),
				method_sig = method_sig)
		txhex,_ = await self.txsign(tx_in,key,from_addr)
		return await self.txsend(txhex)

class Token(TokenCommon):

	def __init__(self,cfg,proto,addr,decimals,rpc=None):
		if type(self).__name__ == 'Token':
			from ...util2 import get_keccak
			self.keccak_256 = get_keccak(cfg)
		self.cfg = cfg
		self.proto = proto
		self.addr = TokenAddr(proto,addr)
		assert isinstance(decimals,int),f'decimals param must be int instance, not {type(decimals)}'
		self.decimals = decimals
		self.base_unit = Decimal('10') ** -self.decimals
		self.rpc = rpc

class ResolvedToken(TokenCommon,metaclass=AsyncInit):

	async def __init__(self,cfg,proto,rpc,addr):
		from ...util2 import get_keccak
		self.keccak_256 = get_keccak(cfg)
		self.cfg = cfg
		self.proto = proto
		self.rpc = rpc
		self.addr = TokenAddr(proto,addr)
		decimals = await self.get_decimals() # requires self.addr!
		if not decimals:
			die( 'TokenNotInBlockchain', f'Token {addr!r} not in blockchain' )
		Token.__init__(self,cfg,proto,addr,decimals,rpc)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txsend: Broadcast a transaction signed by 'mmgen-txsign' to the network
"""

import sys

from .cfg import gc,Config
from .util import async_run, die

opts_data = {
	'sets': [
		('yes', True, 'quiet', True),
		('abort', True, 'autosign', True),
	],
	'text': {
		'desc':    f'Send a signed {gc.proj_name} cryptocoin transaction',
		'usage':   '[opts] [signed transaction file]',
		'options': """
-h, --help      Print this help message
--, --longhelp  Print help message for long options (common options)
-a, --autosign  Send an autosigned transaction created by ‘mmgen-txcreate
                --autosign’.  The removable device is mounted and unmounted
                automatically. The transaction file argument must be omitted
                when using this option
-A, --abort     Abort an unsent transaction created by ‘mmgen-txcreate
                --autosign’ and delete it from the removable device.  The
                transaction may be signed or unsigned.
-d, --outdir= d Specify an alternate directory 'd' for output
-q, --quiet     Suppress warnings; overwrite files without prompting
-s, --status    Get status of a sent transaction (or the current transaction,
                whether sent or unsent, when used with --autosign)
-v, --verbose   Be more verbose
-y, --yes       Answer 'yes' to prompts, suppress non-essential output
"""
	}
}

cfg = Config(opts_data=opts_data)

if cfg.autosign and cfg.outdir:
	die(1, '--outdir cannot be used in combination with --autosign')

if len(cfg._args) == 1:
	infile = cfg._args[0]
	from .fileutil import check_infile
	check_infile(infile)
elif not cfg._args and cfg.autosign:
	from .tx.util import init_removable_device
	from .autosign import Signable
	asi = init_removable_device(cfg)
	asi.do_mount()
	si = Signable.automount_transaction(asi)
	if cfg.abort:
		si.shred_abortable() # prompts user, then raises exception or exits
	elif cfg.status:
		if si.unsent:
			die(1, 'Transaction is unsent')
		if si.unsigned:
			die(1, 'Transaction is unsigned')
	else:
		infile = si.get_unsent()
		cfg._util.qmsg(f'Got signed transaction file ‘{infile}’')
else:
	cfg._opts.usage()

if not cfg.status:
	from .ui import do_license_msg
	do_license_msg(cfg)

async def main():

	from .tx import OnlineSignedTX, SentTX

	if cfg.status and cfg.autosign:
		tx = await si.get_last_created()
	else:
		tx = await OnlineSignedTX(
			cfg        = cfg,
			filename   = infile,
			automount  = cfg.autosign,
			quiet_open = True)

	from .rpc import rpc_init
	tx.rpc = await rpc_init(cfg,tx.proto)

	cfg._util.vmsg(f'Getting {tx.desc} ‘{tx.infile}’')

	if cfg.status:
		if tx.coin_txid:
			cfg._util.qmsg(f'{tx.proto.coin} txid: {tx.coin_txid.hl()}')
		await tx.status.display(usr_req=True)
		sys.exit(0)

	if not cfg.yes:
		tx.info.view_with_prompt('View transaction details?')
		if tx.add_comment(): # edits an existing comment, returns true if changed
			if not cfg.autosign:
				tx.file.write(ask_write_default_yes=True)

	if await tx.send():
		tx2 = await SentTX(cfg=cfg, data=tx.__dict__, automount=cfg.autosign)
		tx2.file.write(
			outdir        = asi.txauto_dir if cfg.autosign else None,
			ask_overwrite = False,
			ask_write     = False)
		tx2.print_contract_addr()

async_run(main())



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
cfg: Configuration classes for the MMGen suite
"""

import sys,os
from collections import namedtuple
from .base_obj import Lockable

def die(*args,**kwargs):
	from .util import die
	die(*args,**kwargs)

def die2(exit_val,s):
	sys.stderr.write(s+'\n')
	sys.exit(exit_val)

class GlobalConstants(Lockable):
	"""
	These values are non-configurable.  They’re constant for a given machine,
	user, executable and MMGen release.
	"""
	_autolock = True

	proj_name          = 'MMGen'
	proj_url           = 'https://github.com/mmgen/mmgen-wallet'
	author             = 'The MMGen Project'
	email              = '<mmgen@tuta.io>'
	Cdates             = '2013-2024'
	dfl_hash_preset    = '3'
	passwd_max_tries   = 5
	min_screen_width   = 80
	min_time_precision = 18

	# must match CoinProtocol.coins
	core_coins = ('btc','bch','ltc','eth','etc','zec','xmr')

	prog_name = os.path.basename(sys.argv[0])
	is_txprog = prog_name == 'mmgen-regtest' or prog_name.startswith('mmgen-tx')

	if sys.platform not in ('linux','win32'):
		die2(1,f'{sys.platform!r}: platform not supported by {proj_name}')

	if os.getenv('HOME'):   # Linux or MSYS2
		home_dir = os.getenv('HOME')
	elif sys.platform == 'win32': # Windows without MSYS2 - not supported
		die2(1,f'$HOME not set!  {proj_name} for Windows must be run in MSYS2 environment')
	else:
		die2(2,'$HOME is not set!  Unable to determine home directory')

	def get_mmgen_data_file(self,filename,package='mmgen'):
		"""
		this is an expensive import, so do only when required
		"""
		# Resource will be unpacked and then cleaned up if necessary, see:
		#    https://docs.python.org/3/library/importlib.html:
		#        Note: This module provides functionality similar to pkg_resources Basic
		#        Resource Access without the performance overhead of that package.
		#    https://importlib-resources.readthedocs.io/en/latest/migration.html
		#    https://setuptools.readthedocs.io/en/latest/pkg_resources.html

		# TODO: remove try..except (workaround for Python 3.9 pylint bug)
		try:
			from importlib.resources import files # Python 3.9 and above
		except ImportError:
			from importlib_resources import files
		return files(package).joinpath('data',filename).read_text()

	@property
	def version(self):
		return self.get_mmgen_data_file(
				filename = 'version',
				package  = 'mmgen_node_tools' if self.prog_name.startswith('mmnode-') else 'mmgen'
			).strip()

	@property
	def release_date(self):
		return self.get_mmgen_data_file(filename='release_date').strip()

gc = GlobalConstants()

class GlobalVars:
	"""
	These are used only by the test suite to redirect msg() and friends to /dev/null
	"""
	stdout = sys.stdout
	stderr = sys.stderr

gv = GlobalVars()

class Config(Lockable):
	"""
	These values are configurable - RHS values are defaults
	Globals are overridden with the following precedence:
	  1 - command line
	  2 - environmental vars
	  3 - config file
	"""
	_autolock = False
	_set_ok = ('usr_randchars','_proto')
	_reset_ok = ('accept_defaults','quiet','verbose','yes')
	_use_class_attr = True
	_default_to_none = True

	# general
	coin        = 'BTC'
	token       = ''
	outdir      = ''
	passwd_file = ''
	network     = 'mainnet'
	testnet     = False
	regtest     = False

	# verbosity / prompting behavior
	quiet           = False
	verbose         = False
	yes             = False
	accept_defaults = False
	no_license      = False

	# limits
	http_timeout       = 60
	usr_randchars      = 30
	fee_adjust         = 1.0
	fee_estimate_confs = 3
	minconf            = 1
	max_tx_file_size   = 100000
	max_input_size     = 1024 * 1024
	min_urandchars     = 10
	max_urandchars     = 80

	# debug
	debug                = False
	debug_daemon         = False
	debug_opts           = False
	debug_rpc            = False
	debug_addrlist       = False
	debug_subseed        = False
	debug_tw             = False
	devtools             = False

	# rpc:
	rpc_host              = ''
	rpc_port              = 0
	rpc_user              = ''
	rpc_password          = ''
	aiohttp_rpc_queue_len = 16
	cached_balances       = False

	# daemons
	daemon_data_dir       = '' # set by user
	daemon_id             = ''
	blacklisted_daemons   = ''
	ignore_daemon_version = False

	# display:
	test_suite_enable_color = False # placeholder
	force_256_color = False
	scroll          = False
	pager           = False
	columns         = 0
	color = bool(
		( sys.stdout.isatty() and not os.getenv('MMGEN_TEST_SUITE_PEXPECT') ) or
		os.getenv('MMGEN_TEST_SUITE_ENABLE_COLOR')
	)

	# miscellaneous features:
	use_internal_keccak_module     = False
	force_standalone_scrypt_module = False
	enable_erigon                  = False
	autochg_ignore_labels          = False

	# regtest:
	bob          = False
	alice        = False
	carol        = False
	regtest_user = ''

	# Monero:
	monero_wallet_rpc_user     = 'monero'
	monero_wallet_rpc_password = ''
	priority                   = 0

	# test suite:
	bogus_send               = False
	bogus_unspent_data       = ''
	debug_utf8               = False
	exec_wrapper             = False
	ignore_test_py_exception = False
	test_suite               = False
	test_suite_autosign_led_simulate = False
	test_suite_autosign_threaded = False
	test_suite_xmr_autosign  = False
	test_suite_cfgtest       = False
	test_suite_deterministic = False
	test_suite_pexpect       = False
	test_suite_popen_spawn   = False
	test_suite_root_pfx      = ''
	hold_protect_disable     = False
	no_daemon_autostart      = False
	names                    = False
	no_timings               = False
	exit_after               = ''
	resuming                 = False
	skipping_deps            = False
	test_datadir             = os.path.join('test','data_dir' + ('','-α')[bool(os.getenv('MMGEN_DEBUG_UTF8'))])

	mnemonic_entry_modes = {}

	# external use:
	_opts  = None
	_proto = None

	# internal use:
	_use_cfg_file           = False
	_use_env                = False

	_forbidden_opts = (
		'data_dir_root',
	)

	_incompatible_opts = (
		('help','longhelp'),
		('bob','alice','carol'),
		('label','keep_label'),
		('tx_id','info'),
		('tx_id','terse_info'),
		('autosign','outdir'),
	)

	_cfg_file_opts = (
		'autochg_ignore_labels',
		'color',
		'daemon_data_dir',
		'debug',
		'fee_adjust',
		'force_256_color',
		'hash_preset',
		'http_timeout',
		'max_input_size',
		'max_tx_file_size',
		'mnemonic_entry_modes',
		'monero_wallet_rpc_password',
		'monero_wallet_rpc_user',
		'no_license',
		'quiet',
		'regtest',
		'rpc_host',
		'rpc_password',
		'rpc_port',
		'rpc_user',
		'scroll',
		'subseeds',
		'testnet',
		'usr_randchars',
		'bch_max_tx_fee',
		'btc_max_tx_fee',
		'eth_max_tx_fee',
		'ltc_max_tx_fee',
		'bch_ignore_daemon_version',
		'btc_ignore_daemon_version',
		'etc_ignore_daemon_version',
		'eth_ignore_daemon_version',
		'ltc_ignore_daemon_version',
		'xmr_ignore_daemon_version',
		'eth_mainnet_chain_names',
		'eth_testnet_chain_names' )

	# Supported environmental vars
	# The corresponding attributes (lowercase, without 'mmgen_') must exist in the class.
	# The 'MMGEN_DISABLE_' prefix sets the corresponding attribute to False.
	_env_opts = (
		'MMGEN_DEBUG_ALL', # special: there is no `debug_all` attribute

		'MMGEN_COLUMNS',
		'MMGEN_TEST_SUITE',
		'MMGEN_TEST_SUITE_AUTOSIGN_LED_SIMULATE',
		'MMGEN_TEST_SUITE_AUTOSIGN_THREADED',
		'MMGEN_TEST_SUITE_XMR_AUTOSIGN',
		'MMGEN_TEST_SUITE_CFGTEST',
		'MMGEN_TEST_SUITE_DETERMINISTIC',
		'MMGEN_TEST_SUITE_ENABLE_COLOR',
		'MMGEN_TEST_SUITE_PEXPECT',
		'MMGEN_TEST_SUITE_POPEN_SPAWN',
		'MMGEN_TEST_SUITE_ROOT_PFX',
		'MMGEN_BLACKLIST_DAEMONS',
		'MMGEN_BOGUS_SEND',
		'MMGEN_BOGUS_UNSPENT_DATA',
		'MMGEN_DEBUG',
		'MMGEN_DEBUG_DAEMON',
		'MMGEN_DEBUG_OPTS',
		'MMGEN_DEBUG_RPC',
		'MMGEN_DEBUG_ADDRLIST',
		'MMGEN_DEBUG_TW',
		'MMGEN_DEBUG_UTF8',
		'MMGEN_DEBUG_SUBSEED',
		'MMGEN_DEVTOOLS',
		'MMGEN_FORCE_256_COLOR',
		'MMGEN_HOLD_PROTECT_DISABLE',
		'MMGEN_HTTP_TIMEOUT',
		'MMGEN_QUIET',
		'MMGEN_NO_LICENSE',
		'MMGEN_RPC_HOST',
		'MMGEN_RPC_FAIL_ON_COMMAND',
		'MMGEN_TESTNET',
		'MMGEN_REGTEST',
		'MMGEN_EXEC_WRAPPER',
		'MMGEN_IGNORE_TEST_PY_EXCEPTION',
		'MMGEN_RPC_BACKEND',
		'MMGEN_IGNORE_DAEMON_VERSION',
		'MMGEN_USE_STANDALONE_SCRYPT_MODULE',
		'MMGEN_ENABLE_ERIGON',
		'MMGEN_DISABLE_COLOR',
	)
	_infile_opts = (
		'keys_from_file',
		'mmgen_keys_from_file',
		'passwd_file',
		'keysforaddrs',
		'comment_file',
		'contract_data',
	)
	# Auto-typechecked and auto-set opts - first value in list is the default
	_ov = namedtuple('autoset_opt_info',['type','choices'])
	_autoset_opts = {
		'fee_estimate_mode': _ov('nocase_pfx', ['conservative','economical']),
		'rpc_backend':       _ov('nocase_pfx', ['auto','httplib','curl','aiohttp','requests']),
	}

	_auto_typeset_opts = {
		'seed_len': int,
		'subseeds': int,
		'vsize_adj': float,
	}

	# test suite:
	err_disp_timeout   = 0.7
	short_disp_timeout = 0.3
	stdin_tty          = sys.stdin.isatty()

	if os.getenv('MMGEN_TEST_SUITE'):
		min_urandchars = 3
		err_disp_timeout = 0.1
		short_disp_timeout = 0.1
		if os.getenv('MMGEN_TEST_SUITE_POPEN_SPAWN'):
			stdin_tty = True
		if gc.prog_name == 'unit_tests.py':
			_set_ok += ('debug_subseed',)
			_reset_ok += ('force_standalone_scrypt_module',)

	if os.getenv('MMGEN_DEBUG_ALL'):
		for name in _env_opts:
			if name[:11] == 'MMGEN_DEBUG':
				os.environ[name] = '1'

	@property
	def data_dir_root(self):
		"""
		location of mmgen.cfg
		"""
		if not hasattr(self,'_data_dir_root'):
			if self._data_dir_root_override:
				self._data_dir_root = os.path.normpath(os.path.abspath(self._data_dir_root_override))
			elif self.test_suite:
				self._data_dir_root = self.test_datadir
			else:
				self._data_dir_root = os.path.join(gc.home_dir,'.'+gc.proj_name.lower())
		return self._data_dir_root

	@property
	def data_dir(self):
		"""
		location of wallet and other data - same as data_dir_root for mainnet
		"""
		if not hasattr(self,'_data_dir'):
			self._data_dir = os.path.normpath(os.path.join(*{
				'regtest': (self.data_dir_root, 'regtest', self.coin.lower(), (self.regtest_user or 'none') ),
				'testnet': (self.data_dir_root, 'testnet'),
				'mainnet': (self.data_dir_root,),
			}[self.network] ))
		return self._data_dir

	def __init__(
			self,
			cfg          = None,
			opts_data    = None,
			init_opts    = None,
			opt_filter   = None,
			parse_only   = False,
			parsed_opts  = None,
			need_proto   = True,
			need_amt     = True,
			do_post_init = False,
			process_opts = False ):

		# Step 1: get user-supplied configuration data from
		#           a) command line, or
		#           b) first argument to constructor;
		#         save to self._uopts:
		self._cloned = {}
		if opts_data or parsed_opts or process_opts:
			assert cfg is None, (
				'Config(): ‘cfg’ cannot be used simultaneously with ' +
				'‘opts_data’, ‘parsed_opts’ or ‘process_opts’' )
			from .opts import UserOpts
			UserOpts(
				cfg          = self,
				opts_data    = opts_data,
				init_opts    = init_opts,
				opt_filter   = opt_filter,
				parse_only   = parse_only,
				parsed_opts  = parsed_opts )
			self._uopt_desc = 'command-line option'
		else:
			if cfg is None:
				self._uopts = {}
			else:
				if '_clone' in cfg:
					assert isinstance( cfg['_clone'], Config )
					self._cloned = cfg['_clone'].__dict__
					for k,v in self._cloned.items():
						if not k.startswith('_'):
							setattr(self,k,v)
					del cfg['_clone']
				self._uopts = cfg
			self._uopt_desc = 'configuration option'

		self._data_dir_root_override = self._cloned.pop(
			'_data_dir_root_override',
			self._uopts.pop('data_dir',None))

		if parse_only and not any(k in self._uopts for k in ['help','longhelp']):
			return

		# Step 2: set cfg from user-supplied data, skipping auto opts; set type from corresponding
		#         class attribute, if it exists:
		auto_opts = tuple(self._autoset_opts) + tuple(self._auto_typeset_opts)
		for key,val in self._uopts.items():
			assert key.isascii() and key.isidentifier() and key[0] != '_', '{key!r}: malformed configuration option'
			assert key not in self._forbidden_opts, '{key!r}: forbidden configuration option'
			if key not in auto_opts:
				if hasattr(type(self), key):
					setattr(
						self,
						key,
						getattr(type(self), key) if val is None else
							conv_type(key, val, getattr(type(self), key), self._uopt_desc))
				elif val is None:
					if hasattr(self, key):
						delattr(self, key)
				else:
					setattr(self, key, val)

		# Step 3: set cfg from environment, skipping already-set opts; save names set from environment:
		self._envopts = tuple(self._set_cfg_from_env()) if self._use_env else ()

		from .term import init_term
		init_term(self) # requires ‘hold_protect_disable’ (set from env)

		from .fileutil import check_or_create_dir
		check_or_create_dir(self.data_dir_root)

		from .util import wrap_ripemd160
		wrap_ripemd160() # ripemd160 required by mmgen_cfg_file() in _set_cfg_from_cfg_file()

		# Step 4: set cfg from cfgfile, skipping already-set opts and auto opts; save set opts and auto
		#         opts to be set:
		# requires ‘data_dir_root’, ‘test_suite_cfgtest’
		self._cfgfile_opts = self._set_cfg_from_cfg_file( self._envopts, need_proto )

		# Step 5: set autoset opts from user-supplied data, cfgfile data, or default values, in that order:
		self._set_autoset_opts( self._cfgfile_opts.autoset )

		# Step 6: set auto typeset opts from user-supplied data or cfgfile data, in that order:
		self._set_auto_typeset_opts( self._cfgfile_opts.auto_typeset )

		if self.regtest or self.bob or self.alice or self.carol or gc.prog_name == 'mmgen-regtest':
			self.network = 'regtest'
			self.regtest_user = 'bob' if self.bob else 'alice' if self.alice else 'carol' if self.carol else None
		else:
			self.network = 'testnet' if self.testnet else 'mainnet'

		self.coin = self.coin.upper()
		self.token = self.token.upper() if self.token else None

		# self.color is finalized, so initialize color:
		if self.color: # MMGEN_DISABLE_COLOR sets this to False
			from .color import init_color
			init_color(num_colors=256 if self.force_256_color else 'auto')

		self._die_on_incompatible_opts()

		check_or_create_dir(self.data_dir)

		if self.debug and gc.prog_name != 'cmdtest.py':
			self.verbose = True
			self.quiet = False

		if self.debug_opts:
			opt_postproc_debug(self)

		from .util import Util
		self._util = Util(self)

		del self._cloned

		self._lock()

		if need_proto:
			from .protocol import warn_trustlevel,init_proto_from_cfg
			warn_trustlevel(self)
			# requires the default-to-none behavior, so do after the lock:
			self._proto = init_proto_from_cfg(self,need_amt=need_amt)

		if self._opts and not do_post_init:
			self._opts.init_bottom(self)

		# Check user-set opts without modifying them
		check_opts(self)

	def _set_cfg_from_env(self):
		for name,val in ((k,v) for k,v in os.environ.items() if k.startswith('MMGEN_')):
			if name == 'MMGEN_DEBUG_ALL':
				continue
			elif name in self._env_opts:
				if val: # ignore empty string values; string value of '0' or 'false' sets variable to False
					disable = name.startswith('MMGEN_DISABLE_')
					gname = name[(6,14)[disable]:].lower()
					if gname in self._uopts: # don’t touch attr if already set by user
						continue
					elif hasattr(self,gname):
						setattr(
							self,
							gname,
							conv_type( name, val, getattr(self,gname), 'environment var', invert_bool=disable ))
						yield gname
					else:
						raise ValueError(f'Name {gname!r} not present in globals')
			else:
				raise ValueError(f'{name!r} is not a valid MMGen environment variable')

	def _set_cfg_from_cfg_file(
			self,
			env_cfg,
			need_proto ):

		_ret = namedtuple('cfgfile_opts',['non_auto','autoset','auto_typeset'])

		if not self._use_cfg_file:
			return _ret( (), {}, {} )

		# check for changes in system template file (term must be initialized)
		from .cfgfile import mmgen_cfg_file
		mmgen_cfg_file(self,'sample')

		ucfg = mmgen_cfg_file(self,'usr')

		self._cfgfile_fn = ucfg.fn

		if need_proto:
			from .protocol import init_proto

		autoset_opts = {}
		auto_typeset_opts = {}
		non_auto_opts = []
		already_set = tuple(self._uopts) + env_cfg

		for d in ucfg.get_lines():
			if d.name in self._cfg_file_opts:
				ns = d.name.split('_')
				if ns[0] in gc.core_coins:
					if not need_proto:
						continue
					nse,tn = (
						(ns[2:],ns[1]=='testnet') if len(ns) > 2 and ns[1] in ('mainnet','testnet') else
						(ns[1:],False)
					)
					# no instance yet, so override _class_ attr:
					cls = init_proto(self, ns[0], tn, need_amt=True, return_cls=True)
					attr = '_'.join(nse)
				else:
					cls = self
					attr = d.name
				refval = getattr(cls,attr)
				val = ucfg.parse_value(d.value,refval)
				if not val:
					die( 'CfgFileParseError', f'Parse error in file {ucfg.fn!r}, line {d.lineno}' )
				val_conv = conv_type( attr, val, refval, 'configuration file option', src=ucfg.fn )
				if not attr in already_set:
					setattr(cls,attr,val_conv)
					non_auto_opts.append(attr)
			elif d.name in self._autoset_opts:
				autoset_opts[d.name] = d.value
			elif d.name in self._auto_typeset_opts:
				auto_typeset_opts[d.name] = d.value
			else:
				die( 'CfgFileParseError', f'{d.name!r}: unrecognized option in {ucfg.fn!r}, line {d.lineno}' )

		return _ret( tuple(non_auto_opts), autoset_opts, auto_typeset_opts )

	def _set_autoset_opts(self,cfgfile_autoset_opts):

		def get_autoset_opt(key,val,src):

			def die_on_err(desc):
				from .util import fmt_list
				die(
					'UserOptError',
					'{a!r}: invalid {b} (not {c}: {d})'.format(
						a = val,
						b = {
							'cmdline': f'parameter for option --{key.replace("_","-")}',
							'cfgfile': f'value for cfg file option {key!r}'
						}[src],
						c = desc,
						d = fmt_list(data.choices) ))

			class opt_type:

				def nocase_str():
					if val.lower() in data.choices:
						return val.lower()
					else:
						die_on_err('one of')

				def nocase_pfx():
					cs = [s for s in data.choices if s.startswith(val.lower())]
					if len(cs) == 1:
						return cs[0]
					else:
						die_on_err('unique substring of')

			data = self._autoset_opts[key]

			return getattr(opt_type,data.type)()

		# Check autoset opts, setting if unset
		for key in self._autoset_opts:

			if key in self._cloned:
				continue

			assert not hasattr(self,key), f'autoset opt {key!r} is already set, but it shouldn’t be!'

			if key in self._uopts:
				val,src = (self._uopts[key],'cmdline')
			elif key in cfgfile_autoset_opts:
				val,src = (cfgfile_autoset_opts[key],'cfgfile')
			else:
				val = None

			if val is None:
				setattr(self, key, self._autoset_opts[key].choices[0])
			else:
				setattr(self, key, get_autoset_opt(key,val,src=src))

	def _set_auto_typeset_opts(self,cfgfile_auto_typeset_opts):

		def do_set(key,val,ref_type):
			assert not hasattr(self,key), f'{key!r} is in cfg!'
			setattr(self,key,None if val is None else ref_type(val))

		for key,ref_type in self._auto_typeset_opts.items():
			if key in self._uopts:
				do_set(key, self._uopts[key], ref_type)
			elif key in cfgfile_auto_typeset_opts:
				do_set(key, cfgfile_auto_typeset_opts[key], ref_type)

	def _post_init(self):
		return self._opts.init_bottom(self)

	def _die_on_incompatible_opts(self):
		for group in self._incompatible_opts:
			bad = [k for k in self.__dict__ if k in group and getattr(self,k) is not None]
			if len(bad) > 1:
				die(1,'Conflicting options: {}'.format(', '.join(map(fmt_opt,bad))))

def check_opts(cfg): # Raises exception if any check fails

	from .util import is_int,Msg

	def get_desc(desc_pfx=''):
		return (
			(desc_pfx + ' ' if desc_pfx else '')
			+ (
				f'parameter for command-line option {fmt_opt(name)!r}'
					if name in cfg._uopts and 'command-line' in cfg._uopt_desc else
				f'value for configuration option {name!r}'
			)
			+ ( ' from environment' if name in cfg._envopts else '')
			+ (f' in {cfg._cfgfile_fn!r}' if name in cfg._cfgfile_opts.non_auto else '')
		)

	def display_opt(name,val='',beg='For selected',end=':\n'):
		from .util import msg_r
		msg_r('{} option {!r}{}'.format(
			beg,
			f'{fmt_opt(name)}={val}' if val else fmt_opt(name),
			end ))

	def opt_compares(val,op_str,target):
		import operator
		if not {
			'<':  operator.lt,
			'<=': operator.le,
			'>':  operator.gt,
			'>=': operator.ge,
			'=':  operator.eq,
		}[op_str](val,target):
			die( 'UserOptError', f'{val}: invalid {get_desc()} (not {op_str} {target})' )

	def opt_is_int(val,desc_pfx=''):
		if not is_int(val):
			die( 'UserOptError', f'{val!r}: invalid {get_desc(desc_pfx)} (not an integer)' )

	def opt_is_in_list(val,tlist,desc_pfx=''):
		if val not in tlist:
			q,sep = (('',','),("'","','"))[isinstance(tlist[0],str)]
			die( 'UserOptError', '{q}{v}{q}: invalid {w}\nValid choices: {q}{o}{q}'.format(
				v = val,
				w = get_desc(desc_pfx),
				q = q,
				o = sep.join(map(str,sorted(tlist))) ))

	def opt_unrecognized():
		die( 'UserOptError', f'{val!r}: unrecognized {get_desc()}' )

	class check_funcs:

		def in_fmt():
			from .wallet import get_wallet_data
			wd = get_wallet_data(fmt_code=val)
			if not wd:
				opt_unrecognized()
			if name == 'out_fmt':
				p = 'hidden_incog_output_params'

				if wd.type == 'incog_hidden' and not getattr(cfg,p):
					die( 'UserOptError',
						'Hidden incog format output requested.  ' +
						f'You must supply a file and offset with the {fmt_opt(p)!r} option' )

				if wd.base_type == 'incog_base' and cfg.old_incog_fmt:
					display_opt(name,val,beg='Selected',end=' ')
					display_opt('old_incog_fmt',beg='conflicts with',end=':\n')
					die( 'UserOptError', 'Export to old incog wallet format unsupported' )
				elif wd.type == 'brain':
					die( 'UserOptError', 'Output to brainwallet format unsupported' )

		out_fmt = in_fmt

		def hidden_incog_input_params():
			a = val.rsplit(',',1) # permit comma in filename
			if len(a) != 2:
				display_opt(name,val)
				die( 'UserOptError', 'Option requires two comma-separated arguments' )

			fn,offset = a
			opt_is_int(offset)

			from .fileutil import check_infile,check_outdir,check_outfile
			if name == 'hidden_incog_input_params':
				check_infile(fn,blkdev_ok=True)
				key2 = 'in_fmt'
			else:
				try:
					os.stat(fn)
				except:
					b = os.path.dirname(fn)
					if b:
						check_outdir(b)
				else:
					check_outfile(fn,blkdev_ok=True)
				key2 = 'out_fmt'

			if hasattr(cfg,key2):
				val2 = getattr(cfg,key2)
				from .wallet import get_wallet_data
				wd = get_wallet_data('incog_hidden')
				if val2 and val2 not in wd.fmt_codes:
					die( 'UserOptError', f'Option conflict:\n  {fmt_opt(name)}, with\n  {fmt_opt(key2)}={val2}' )

		hidden_incog_output_params = hidden_incog_input_params

		def subseeds():
			from .subseed import SubSeedIdxRange
			opt_compares(val,'>=',SubSeedIdxRange.min_idx)
			opt_compares(val,'<=',SubSeedIdxRange.max_idx)

		def seed_len():
			from .seed import Seed
			opt_is_in_list(int(val),Seed.lens)

		def hash_preset():
			from .crypto import Crypto
			opt_is_in_list(val,list(Crypto.hash_presets.keys()))

		def brain_params():
			a = val.split(',')
			if len(a) != 2:
				display_opt(name,val)
				die( 'UserOptError', 'Option requires two comma-separated arguments' )

			opt_is_int( a[0], desc_pfx='seed length' )
			from .seed import Seed
			opt_is_in_list( int(a[0]), Seed.lens, desc_pfx='seed length' )

			from .crypto import Crypto
			opt_is_in_list( a[1], list(Crypto.hash_presets.keys()), desc_pfx='hash preset' )

		def usr_randchars():
			if val != 0:
				opt_compares(val,'>=',cfg.min_urandchars)
				opt_compares(val,'<=',cfg.max_urandchars)

		def tx_confs():
			opt_is_int(val)
			opt_compares(int(val),'>=',1)

		def vsize_adj():
			from .util import ymsg
			ymsg(f'Adjusting transaction vsize by a factor of {val:1.2f}')

		def daemon_id():
			from .daemon import CoinDaemon
			opt_is_in_list(val,CoinDaemon.all_daemon_ids())

		def locktime():
			opt_is_int(val)
			opt_compares(int(val),'>',0)

		def columns():
			opt_compares(val,'>',10)

	# TODO: add checks for token, rbf, tx_fee
	check_funcs_names = tuple(check_funcs.__dict__)
	for name in tuple(cfg._uopts) + cfg._envopts + cfg._cfgfile_opts.non_auto:
		val = getattr(cfg,name)
		if name in cfg._infile_opts:
			from .fileutil import check_infile
			check_infile(val) # file exists and is readable - dies on error
		elif name == 'outdir':
			from .fileutil import check_outdir
			check_outdir(val) # dies on error
		elif name in check_funcs_names:
			getattr(check_funcs,name)()
		elif cfg.debug:
			Msg(f'check_opts(): No test for config opt {name!r}')

def fmt_opt(o):
	return '--' + o.replace('_','-')

def opt_postproc_debug(cfg):
	none_opts = [k for k in dir(cfg) if k[:2] != '__' and getattr(cfg,k) is None]
	from .util import Msg
	Msg('\n    Configuration opts:')
	for e in [d for d in dir(cfg) if d[:2] != '__']:
		Msg(f'        {e:<20}: {getattr(cfg,e)}')
	Msg("    Configuration opts set to 'None':")
	Msg('        {}\n'.format('\n        '.join(none_opts)))
	Msg('\n=== end opts.py debug ===\n')

def conv_type(
		name,
		val,
		refval,
		desc,
		invert_bool = False,
		src         = None ):

	def do_fail():
		die(1,'{a!r}: invalid value for {b} {c!r}{d} (must be of type {e!r})'.format(
			a = val,
			b = desc,
			c = fmt_opt(name) if 'command-line' in desc else name,
			d = f' in {src!r}' if src else '',
			e = type(refval).__name__ ))

	if type(refval) is bool:
		v = str(val).lower()
		ret = (
			True  if v in ('true','yes','1','on') else
			False if v in ('false','no','none','0','off','') else
			None
		)
		return do_fail() if ret is None else (not ret) if invert_bool else ret
	else:
		try:
			return type(refval)(not val if invert_bool else val)
		except:
			do_fail()



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
util2: Less frequently-used variables, classes and utility functions for the MMGen suite
"""

import sys,re,time
from .util import msg,suf,hexdigits,die

def die_wait(delay,ev=0,s=''):
	assert isinstance(delay,int)
	assert isinstance(ev,int)
	if s:
		msg(s)
	time.sleep(delay)
	sys.exit(ev)

def die_pause(ev=0,s=''):
	assert isinstance(ev,int)
	if s:
		msg(s)
	input('Press ENTER to exit')
	sys.exit(ev)

# monkey-patch function for monero-python: permits its use with pycryptodome (e.g. MSYS2)
# instead of the expected pycryptodomex
def load_cryptodomex():
	try:
		import Cryptodome # cryptodomex
	except ImportError:
		try:
			import Crypto # cryptodome
		except ImportError:
			die(2,'Unable to import either the ‘pycryptodomex’ or ‘pycryptodome’ package')
		else:
			sys.modules['Cryptodome'] = Crypto

# called with no arguments by pyethereum.utils:
def get_keccak(cfg=None,cached_ret=[]):

	if not cached_ret:
		if cfg and cfg.use_internal_keccak_module:
			cfg._util.qmsg('Using internal keccak module by user request')
			from .contrib.keccak import keccak_256
		else:
			try:
				from Cryptodome.Hash import keccak
			except ImportError as e:
				try:
					from Crypto.Hash import keccak
				except ImportError as e2:
					msg(f'{e2} and {e}')
					die('MMGenImportError',
						'Please install the ‘pycryptodome’ or ‘pycryptodomex’ package on your system')
			def keccak_256(data):
				return keccak.new(data=data,digest_bytes=32)
		cached_ret.append(keccak_256)

	return cached_ret[0]

# From 'man dd':
# c=1, w=2, b=512, kB=1000, K=1024, MB=1000*1000, M=1024*1024,
# GB=1000*1000*1000, G=1024*1024*1024, and so on for T, P, E, Z, Y.
bytespec_map = (
	('c',  1),
	('w',  2),
	('b',  512),
	('kB', 1000),
	('K',  1024),
	('MB', 1000000),
	('M',  1048576),
	('GB', 1000000000),
	('G',  1073741824),
	('TB', 1000000000000),
	('T',  1099511627776),
	('PB', 1000000000000000),
	('P',  1125899906842624),
	('EB', 1000000000000000000),
	('E',  1152921504606846976),
)

def int2bytespec(n,spec,fmt,print_sym=True,strip=False,add_space=False):

	def spec2int(spec):
		for k,v in bytespec_map:
			if k == spec:
				return v
		else:
			die(1,f'{spec!r}: unrecognized bytespec')

	ret = f'{n/spec2int(spec):{fmt}f}'
	if strip:
		ret = ret.rstrip('0')
		return (
			ret
			+ ('0' if ret.endswith('.') else '')
			+ ((' ' if add_space else '') + spec if print_sym else '') )
	else:
		return (
			ret
			+ ((' ' if add_space else '') + spec if print_sym else '') )

def parse_bytespec(nbytes):
	m = re.match(r'([0123456789.]+)(.*)',nbytes)
	if m:
		if m.group(2):
			for k,v in bytespec_map:
				if k == m.group(2):
					from decimal import Decimal
					return int(Decimal(m.group(1)) * v)
			else:
				msg("Valid byte specifiers: '{}'".format("' '".join([i[0] for i in bytespec_map])))
		elif '.' in nbytes:
			raise ValueError('fractional bytes not allowed')
		else:
			return int(nbytes)

	die(1,f'{nbytes!r}: invalid byte specifier')

def format_elapsed_days_hr(t,now=None,cached={}):
	e = int((now or time.time()) - t)
	if not e in cached:
		days = abs(e) // 86400
		cached[e] = f'{days} day{suf(days)} ' + ('ago' if e > 0 else 'in the future')
	return cached[e]

def format_elapsed_hr(t, now=None, cached={}, rel_now=True, show_secs=False):
	e = int((now or time.time()) - t)
	key = f'{e}:{rel_now}:{show_secs}'
	if not key in cached:
		def add_suffix():
			return (
				((' ago'           if rel_now else '') if e > 0 else
				(' in the future' if rel_now else ' (negative elapsed)'))
					if (abs_e if show_secs else abs_e // 60) else
				('just now' if rel_now else ('0 ' + ('seconds' if show_secs else 'minutes')))
			)
		abs_e = abs(e)
		data = (
			('day',    abs_e // 86400),
			('hour',   abs_e // 3600 % 24),
			('minute', abs_e // 60 % 60),
			('second', abs_e % 60),
		) if show_secs else (
			('day',    abs_e // 86400),
			('hour',   abs_e // 3600 % 24),
			('minute', abs_e // 60 % 60),
		)
		cached[key] = ' '.join(f'{n} {desc}{suf(n)}' for desc, n in data if n) + add_suffix()
	return cached[key]

def pretty_format(s,width=80,pfx=''):
	out = []
	while s:
		if len(s) <= width:
			out.append(s)
			break
		i = s[:width].rfind(' ')
		out.append(s[:i])
		s = s[i+1:]
	return pfx + ('\n'+pfx).join(out)

def block_format(data,gw=2,cols=8,line_nums=None,data_is_hex=False):
	assert line_nums in (None,'hex','dec'),"'line_nums' must be one of None, 'hex' or 'dec'"
	ln_fs = '{:06x}: ' if line_nums == 'hex' else '{:06}: '
	bytes_per_chunk = gw
	if data_is_hex:
		gw *= 2
	nchunks = len(data)//gw + bool(len(data)%gw)
	return ''.join(
		('' if (line_nums is None or i % cols) else ln_fs.format(i*bytes_per_chunk))
		+ data[i*gw:i*gw+gw]
		+ (' ' if (not cols or (i+1) % cols) else '\n')
			for i in range(nchunks)
	).rstrip() + '\n'

def pretty_hexdump(data,gw=2,cols=8,line_nums=None):
	return block_format(data.hex(),gw,cols,line_nums,data_is_hex=True)

def decode_pretty_hexdump(data):
	pat = re.compile(fr'^[{hexdigits}]+:\s+')
	lines = [pat.sub('',line) for line in data.splitlines()]
	try:
		return bytes.fromhex(''.join((''.join(lines).split())))
	except:
		msg('Data not in hexdump format')
		return False



#!/usr/bin/env python3

"""
test.unit_tests_d.ut_baseconv: Base conversion unit test for the MMGen suite
"""

from mmgen.util import msg,msg_r

from ..include.common import cfg,qmsg,qmsg_r,vmsg,vmsg_r

class unit_test:

	vectors = {
		'b58': (
			(('00',None),'1'),
			(('00',1),'1'),
			(('00',2),'11'),
			(('01',None),'2'),
			(('01',1),'2'),
			(('01',2),'12'),
			(('0f',None),'G'),
			(('0f',1),'G'),
			(('0f',2),'1G'),
			(('deadbeef',None),'6h8cQN'),
			(('deadbeef',20),'111111111111116h8cQN'),
			(('00000000',None),'1'),
			(('00000000',20),'11111111111111111111'),
			(('ffffffff',None),'7YXq9G'),
			(('ffffffff',20),'111111111111117YXq9G'),
			(('ff'*16,'seed'),'YcVfxkQb6JRzqk5kF2tNLv'),
			(('ff'*24,'seed'),'QLbz7JHiBTspS962RLKV8GndWFwiEaqKL'),
			(('ff'*32,'seed'),'JEKNVnkbo3jma5nREBBJCDoXFVeKkD56V3xKrvRmWxFG'),
			(('00'*16,'seed'),'1111111111111111111111'),
			(('00'*24,'seed'),'111111111111111111111111111111111'),
			(('00'*32,'seed'),'11111111111111111111111111111111111111111111'),
		),
		# MMGen-flavored base32 using simple base conversion
		'b32': (
			(('00',None),'A'),
			(('00',1),'A'),
			(('00',2),'AA'),
			(('01',None),'B'),
			(('01',1),'B'),
			(('01',2),'AB'),
			(('0f',None),'P'),
			(('0f',1),'P'),
			(('0f',2),'AP'),
			(('deadbeef',None),'DPK3PXP'),
			(('deadbeef',20),'AAAAAAAAAAAAADPK3PXP'),
			(('00000000',None),'A'),
			(('00000000',20),'AAAAAAAAAAAAAAAAAAAA'),
			(('ffffffff',None),'D777777'),
			(('ffffffff',20),'AAAAAAAAAAAAAD777777'),
		),
		'b16': (
			(('00',None),'0'),
			(('00',1),'0'),
			(('00',2),'00'),
			(('01',None),'1'),
			(('01',1),'1'),
			(('01',2),'01'),
			(('0f',None),'f'),
			(('0f',1),'f'),
			(('0f',2),'0f'),
			(('deadbeef',None),'deadbeef'),
			(('deadbeef',20),'000000000000deadbeef'),
			(('00000000',None),'0'),
			(('00000000',20),'00000000000000000000'),
			(('ffffffff',None),'ffffffff'),
			(('ffffffff',20),'000000000000ffffffff'),
		),
		'b10': (
			(('00',None),'0'),
			(('00',1),'0'),
			(('00',2),'00'),
			(('01',None),'1'),
			(('01',1),'1'),
			(('01',2),'01'),
			(('0f',None),'15'),
			(('0f',1),'15'),
			(('0f',2),'15'),
			(('deadbeef',None),'3735928559'),
			(('deadbeef',20),'00000000003735928559'),
			(('00000000',None),'0'),
			(('00000000',20),'00000000000000000000'),
			(('ffffffff',None),'4294967295'),
			(('ffffffff',20),'00000000004294967295'),
		),
		'b8': (
			(('00',None),'0'),
			(('00',1),'0'),
			(('00',2),'00'),
			(('01',None),'1'),
			(('01',1),'1'),
			(('01',2),'01'),
			(('0f',None),'17'),
			(('0f',1),'17'),
			(('0f',2),'17'),
			(('deadbeef',None),'33653337357'),
			(('deadbeef',20),'00000000033653337357'),
			(('00000000',None),'0'),
			(('00000000',20),'00000000000000000000'),
			(('ffffffff',None),'37777777777'),
			(('ffffffff',20),'00000000037777777777'),
		),
		'b6d': (
			(('00',None),'1'),
			(('00',1),'1'),
			(('00',2),'11'),
			(('01',None),'2'),
			(('01',1),'2'),
			(('01',2),'12'),
			(('0f',None),'34'),
			(('0f',1),'34'),
			(('0f',2),'34'),
			(('010f',None),'2242'),
			(('010f',20),'11111111111111112242'),
			(('deadbeef',None),'2525524636426'),
			(('deadbeef',20),'11111112525524636426'),
			(('00000000',None),'1'),
			(('00000000',20),'11111111111111111111'),
			(('ffffffff',None),'2661215126614'),
			(('ffffffff',20),'11111112661215126614'),

			(('ff'*16,'seed'),'34164464641266661652465154654653354436664555521414'),
			(('ff'*24,'seed'),'246111411433323364222465566552324652566121541623426135163525451613554313654'),
			(('ff'*32,'seed'),'2132521653312613134145131423465414636252114131225324246311141642456513416322412146151432142242565134'),
			(('00'*16,'seed'),'1'*50),
			(('00'*24,'seed'),'1'*75),
			(('00'*32,'seed'),'1'*100),
		),
		'mmgen': (
			(('deadbeefdeadbeefdeadbeefdeadbeef','seed'),
			'table cast forgive master funny gaze sadness ripple million paint moral match' ),
			(('deadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef','seed'),
			'swirl maybe anymore mix scale stray fog use approach page crime rhyme ' +
			'class former strange window snap soon'),
			(('deadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef','seed'),
			'swell type milk figure cheese phone fill black test bloom heard comfort ' +
			'image terrible radio lesson own reply battle goal goodbye need laugh stream'),
			(('ffffffffffffffffffffffffffffffff','seed'),
			'yellow yeah show bowl season spider cling defeat poison law shelter reflect'),
			(('ffffffffffffffffffffffffffffffffffffffffffffffff','seed'),
			'yeah youth quit fail perhaps drum out person young click skin ' +
			'weird inside silently perfectly together anyone memory'),
			(('ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff','seed'),
			'wrote affection object cell opinion here laughter stare honest north cost begin ' +
			'murder something yourself effort acid dot doubt game broke tell guilt innocent'),
			(('00000000000000000000000000000001','seed'),
			'able ' * 11 + 'about'),
			(('000000000000000000000000000000000000000000000001','seed'),
			'able ' * 17 + 'about'),
			(('0000000000000000000000000000000000000000000000000000000000000001','seed'),
			'able ' * 23 + 'about'),
		),
	}

	def run_test(self,name,ut):

		msg_r('Testing base conversion routines...')

		from mmgen.baseconv import baseconv
		perr = "length of {!r} less than pad length ({})"
		rerr = "return value ({!r}) does not match reference value ({!r})"

		qmsg_r('\nChecking hex-to-base conversion:')
		for base,data in self.vectors.items():
			fs = "  {h:%s}  {p:<6} {r}" % max(len(d[0][0]) for d in data)
			if not cfg.verbose:
				qmsg_r(f' {base}')
			vmsg(f'\nBase: {base}')
			vmsg(fs.format(h='Input',p='Pad',r='Output'))
			for (hexstr,pad),ret_chk in data:
				ret = baseconv(base).fromhex(hexstr,pad=pad,tostr=True)
				if pad != 'seed':
					assert len(ret) >= (pad or 0), perr.format(ret,pad or 0)
				assert ret == ret_chk, rerr.format(ret,ret_chk)
				vmsg(fs.format(h=hexstr,r=ret,p=str(pad)))
#				msg("(('{h}',{p}),'{r}'),".format(h=hexstr,r=ret,c=ret_chk,p=pad))
#			msg('')
#		return True
		qmsg_r('\nChecking base-to-hex conversion:')
		for base,data in self.vectors.items():
			fs = "  {h:%s}  {p:<6} {r}" % max(len(d[1]) for d in data)
			if not cfg.verbose:
				qmsg_r(f' {base}')
			vmsg(f'\nBase: {base}')
			vmsg(fs.format(h='Input',p='Pad',r='Output'))
			for (hexstr,pad),ret_chk in data:
				if type(pad) is int:
					pad = len(hexstr)
				ret = baseconv(base).tohex( ret_chk.split() if base == 'mmgen' else ret_chk, pad=pad )
				if pad is None:
					assert int(ret,16) == int(hexstr,16), rerr.format(int(ret,16),int(hexstr,16))
				else:
					assert ret == hexstr, rerr.format(ret,hexstr)
				vmsg(fs.format(h=ret_chk,r=ret,p=str(pad)))
#				msg("(('{h}',{p}),'{r}'),".format(h=hexstr,r=ret_chk,c=ret_chk,p=pad))

		qmsg_r('\nChecking wordlist checksums:')
		vmsg('')

		for wl_id in baseconv.constants['wl_chksum']:
			vmsg_r(f'  {wl_id+":":9}')
			baseconv(wl_id).check_wordlist(cfg)

		qmsg('')

		qmsg('Checking error handling:')

		bad_b58 = 'I'*22
		bad_b58len = 'a'*23

		fr58 = baseconv('b58').fromhex
		to58 = baseconv('b58').tohex
		to32 = baseconv('b32').tohex
		to8  = baseconv('b8').tohex

		hse = 'HexadecimalStringError'
		bce = 'BaseConversionError'
		bpe = 'BaseConversionPadError'
		sle = 'SeedLengthError'

		bad_data = (
			('hexstr',           hse, ': not a hexadecimal str', lambda:fr58('x')),
			('hexstr (seed)',    hse, 'seed data not a hexadec', lambda:fr58('x',pad='seed')),
			('hexstr (empty)',   bce, 'empty data not allowed',  lambda:fr58('')),
			('b58 data',         bce, ': not in base58',         lambda:to58('IfFzZ')),
			('b58 data (seed)',  bce, 'seed data not in base58', lambda:to58(bad_b58,pad='seed')),
			('b58 len (seed)',   bce, 'invalid length for',      lambda:to58(bad_b58len,pad='seed')),
			('b58 data (empty)', bce, 'empty base58 data',       lambda:to58('')),
			('b8 data (empty)' , bce, 'empty base8 string data', lambda:to8('')),
			('b32 data',         bce, 'not in MMGen base32',     lambda:to32('1az')),
			('pad arg (in)',     bpe, "illegal value for 'pad'", lambda:fr58('ff',pad='foo')),
			('pad arg (in)',     bpe, "illegal value for 'pad'", lambda:fr58('ff',pad=False)),
			('pad arg (in)',     bpe, "illegal value for 'pad'", lambda:fr58('ff',pad=True)),
			('seedlen (in)',     sle, 'invalid byte length',     lambda:fr58('ff',pad='seed')),
			('pad arg (out)',    bpe, "illegal value for 'pad'", lambda:to58('Z',pad='foo')),
			('pad arg (out)',    bpe, "illegal value for 'pad'", lambda:to58('Z',pad=False)),
			('pad arg (out)',    bpe, "illegal value for 'pad'", lambda:to58('Z',pad=True)),
			('seedlen (out)',    bce, 'invalid length for seed', lambda:to58('Z',pad='seed')),
		)

		ut.process_bad_data(bad_data)

		msg('OK')

		return True



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_py_d.ct_chainsplit: Forking scenario tests for the cmdtest.py test suite
This module is unmaintained and currently non-functional
"""

from mmgen.util import die

from .common import get_file_with_ext,rt_pw
from .ct_regtest import CmdTestRegtest

class CmdTestChainsplit(CmdTestRegtest):
	'forking scenario tests for the cmdtest.py test suite'
	cmd_group = (
		('split_setup',        'regtest forking scenario setup'),
		('walletgen_bob',      "generating Bob's wallet"),
		('addrgen_bob',        "generating Bob's addresses"),
		('addrimport_bob',     "importing Bob's addresses"),
		('fund_bob',           "funding Bob's wallet"),
		('split_fork',         'regtest split fork'),
		('split_start_btc',    'start regtest daemon (BTC)'),
		('split_start_b2x',    'start regtest daemon (B2X)'),
		('split_gen_btc',      'mining a block (BTC)'),
		('split_gen_b2x',      'mining 100 blocks (B2X)'),
		('split_do_split',     'creating coin splitting transactions'),
		('split_sign_b2x',     'signing B2X split transaction'),
		('split_sign_btc',     'signing BTC split transaction'),
		('split_send_b2x',     'sending B2X split transaction'),
		('split_send_btc',     'sending BTC split transaction'),
		('split_gen_btc',      'mining a block (BTC)'),
		('split_gen_b2x2',     'mining a block (B2X)'),
		('split_txdo_timelock_bad_btc', 'sending transaction with bad locktime (BTC)'),
		('split_txdo_timelock_good_btc','sending transaction with good locktime (BTC)'),
		('split_txdo_timelock_bad_b2x', 'sending transaction with bad locktime (B2X)'),
		('split_txdo_timelock_good_b2x','sending transaction with good locktime (B2X)'),
	)

	def split_setup(self):
		if self.proto.coin != 'BTC':
			die(1,'Test valid only for coin BTC')
		self.coin = 'BTC'
		return self.setup()

	def split_fork(self):
		self.coin = 'B2X'
		t = self.spawn('mmgen-regtest',['fork','btc'])
		t.expect('Creating fork from coin')
		t.expect('successfully created')
		t.ok()

	def split_start(self,coin):
		self.coin = coin
		t = self.spawn('mmgen-regtest',['bob'])
		t.expect('Starting')
		t.expect('done')
		t.ok()

	def split_start_btc(self):
		self.regtest_start(coin='BTC')

	def split_start_b2x(self):
		self.regtest_start(coin='B2X')

	def split_gen_btc(self):
		self.regtest_generate(coin='BTC')

	def split_gen_b2x(self):
		self.regtest_generate(coin='B2X',num_blocks=100)

	def split_gen_b2x2(self):
		self.regtest_generate(coin='B2X')

	def split_do_split(self):
		self.coin = 'B2X'
		sid = self.regtest_user_sid('bob')
		t = self.spawn('mmgen-split',[
			'--bob',
			'--outdir='+self.tmpdir,
			'--tx-fees=0.0001,0.0003',
			sid+':S:1',sid+':S:2'])
		t.expect(r'\[q\]uit menu, .*?:.','q', regex=True)
		t.expect('outputs to spend: ','1\n')

		for tx in ('timelocked','split'):
			for q in ('fee','change'):
				t.expect('OK? (Y/n): ','y')
			t.do_comment(False)
			t.view_tx('t')

		t.written_to_file('Long chain (timelocked) transaction')
		t.written_to_file('Short chain transaction')
		t.ok()

	def split_sign(self,coin,ext):
		wf = get_file_with_ext(self.regtest_user_dir('bob',coin=coin.lower()),'mmdat')
		txfile = self.get_file_with_ext(ext,no_dot=True)
		self.coin = coin
		self.txsign(txfile,wf,extra_opts=['--bob'])

	def split_sign_b2x(self):
		return self.regtest_sign(coin='B2X',ext='533].rawtx')

	def split_sign_btc(self):
		return self.regtest_sign(coin='BTC',ext='9997].rawtx')

	def split_send(self,coin,ext):
		self.coin = coin
		txfile = self.get_file_with_ext(ext,no_dot=True)
		self.txsend(txfile,bogus_send=False,extra_opts=['--bob'])

	def split_send_b2x(self):
		return self.regtest_send(coin='B2X',ext='533].sigtx')

	def split_send_btc(self):
		return self.regtest_send(coin='BTC',ext='9997].sigtx')

	def split_txdo_timelock(self,coin,locktime,bad_locktime):
		self.coin = coin
		sid = self.regtest_user_sid('bob')
		self.regtest_user_txdo(
			'bob',
			'0.0001',
			[sid+':S:5'],
			'1',
			pw           = rt_pw,
			extra_args   = ['--locktime='+str(locktime)],
			bad_locktime = bad_locktime)

	def split_txdo_timelock_bad_btc(self):
		self.regtest_txdo_timelock('BTC',locktime=8888,bad_locktime=True)
	def split_txdo_timelock_good_btc(self):
		self.regtest_txdo_timelock('BTC',locktime=1321009871,bad_locktime=False)
	def split_txdo_timelock_bad_b2x(self):
		self.regtest_txdo_timelock('B2X',locktime=8888,bad_locktime=True)
	def split_txdo_timelock_good_b2x(self):
		self.regtest_txdo_timelock('B2X',locktime=1321009871,bad_locktime=False)



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.addrgen: Ethereum address generation class for the MMGen suite
"""

from ...addrgen import addr_generator,check_data
from ...addr import CoinAddr

class ethereum(addr_generator.keccak):

	@check_data
	def to_addr(self,data):
		return CoinAddr(
			self.proto,
			self.keccak_256(data.pubkey[1:]).hexdigest()[24:] )



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mmgen-txdo: Create, sign and broadcast an online MMGen transaction
"""

from .cfg import gc,Config
from .util import die,fmt_list,async_run
from .subseed import SubSeedIdxRange

opts_data = {
	'sets': [('yes', True, 'quiet', True)],
	'text': {
		'desc': f'Create, sign and send an {gc.proj_name} transaction',
		'usage':   '[opts]  [<addr,amt> ...] <change addr, addrlist ID or addr type> [addr file ...] ' +
					'[seed source ...]',
		'options': """
-h, --help             Print this help message
--, --longhelp         Print help message for long options (common options)
-A, --fee-adjust=    f Adjust transaction fee by factor 'f' (see below)
-b, --brain-params=l,p Use seed length 'l' and hash preset 'p' for
                       brainwallet input
-B, --no-blank         Don't blank screen before displaying unspent outputs
-c, --comment-file=  f Source the transaction's comment from file 'f'
-C, --fee-estimate-confs=c Desired number of confirmations for fee estimation
                       (default: {cfg.fee_estimate_confs})
-d, --outdir=        d Specify an alternate directory 'd' for output
-D, --contract-data= D Path to hex-encoded contract data (ETH only)
-e, --echo-passphrase  Print passphrase to screen when typing it
-E, --fee-estimate-mode=M Specify the network fee estimate mode.  Choices:
                       {fe_all}.  Default: {fe_dfl!r}
-f, --fee=           f Transaction fee, as a decimal {cu} amount or as
                       {fu} (an integer followed by {fl!r}).
                       See FEE SPECIFICATION below.  If omitted, fee will be
                       calculated using network fee estimation.
-g, --gas=           g Specify start gas amount in Wei (ETH only)
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated)
-i, --in-fmt=        f Input is from wallet format 'f' (see FMT CODES below)
-I, --inputs=        i Specify transaction inputs (comma-separated list of
                       MMGen IDs or coin addresses).  Note that ALL unspent
                       outputs associated with each address will be included.
-l, --seed-len=      l Specify wallet seed length of 'l' bits. This option
                       is required only for brainwallet and incognito inputs
                       with non-standard (< {dsl}-bit) seed lengths.
-k, --keys-from-file=f Provide additional keys for non-{pnm} addresses
-K, --keygen-backend=n Use backend 'n' for public key generation.  Options
                       for {coin_id}: {kgs}
-l, --locktime=      t Lock time (block height or unix seconds) (default: 0)
-L, --autochg-ignore-labels Ignore labels when autoselecting change addresses
-m, --minconf=n        Minimum number of confirmations required to spend
                       outputs (default: 1)
-M, --mmgen-keys-from-file=f Provide keys for {pnm} addresses in a key-
                       address file (output of '{pnl}-keygen'). Permits
                       online signing without an {pnm} seed source. The
                       key-address file is also used to verify {pnm}-to-{cu}
                       mappings, so the user should record its checksum.
-O, --old-incog-fmt    Specify old-format incognito input
-p, --hash-preset=   p Use the scrypt hash parameters defined by preset 'p'
                       for password hashing (default: '{gc.dfl_hash_preset}')
-P, --passwd-file=   f Get {pnm} wallet passphrase from file 'f'
-R, --no-rbf           Make transaction non-replaceable (non-replace-by-fee
                       according to BIP 125)
-q, --quiet            Suppress warnings; overwrite files without prompting
-u, --subseeds=      n The number of subseed pairs to scan for (default: {ss},
                       maximum: {ss_max}). Only the default or first supplied
                       wallet is scanned for subseeds.
-v, --verbose          Produce more verbose output
-V, --vsize-adj=     f Adjust transaction's estimated vsize by factor 'f'
-X, --cached-balances  Use cached balances (Ethereum only)
-y, --yes              Answer 'yes' to prompts, suppress non-essential output
-z, --show-hash-presets Show information on available hash presets
""",
	'notes': """
{c}\n{F}

                                 SIGNING NOTES
{s}
Seed source files must have the canonical extensions listed in the 'FileExt'
column below:

FMT CODES:

  {f}

{x}"""
	},
	'code': {
		'options': lambda cfg,proto,help_notes,s: s.format(
			gc=gc,cfg=cfg,pnm=gc.proj_name,pnl=gc.proj_name.lower(),
			kgs=help_notes('keygen_backends'),
			coin_id=help_notes('coin_id'),
			fu=help_notes('rel_fee_desc'),
			fl=help_notes('fee_spec_letters'),
			ss=help_notes('dfl_subseeds'),
			ss_max=SubSeedIdxRange.max_idx,
			fe_all=fmt_list(cfg._autoset_opts['fee_estimate_mode'].choices,fmt='no_spc'),
			fe_dfl=cfg._autoset_opts['fee_estimate_mode'].choices[0],
			dsl=help_notes('dfl_seed_len'),
			cu=proto.coin),
		'notes': lambda cfg,help_notes,s: s.format(
			c = help_notes('txcreate'),
			F = help_notes('fee'),
			s = help_notes('txsign'),
			f = help_notes('fmt_codes'),
			x = help_notes('txcreate_examples') ),
	}
}

cfg = Config(opts_data=opts_data)

from .tx import NewTX, SentTX
from .tx.sign import txsign,get_seed_files,get_keyaddrlist,get_keylist

seed_files = get_seed_files(cfg,cfg._args)

async def main():

	tx1 = await NewTX(cfg=cfg,proto=cfg._proto)

	from .rpc import rpc_init
	tx1.rpc = await rpc_init(cfg)

	tx2 = await tx1.create(
		cmd_args = cfg._args,
		locktime = int(cfg.locktime or 0),
		caller   = 'txdo' )

	kal = get_keyaddrlist(cfg,cfg._proto)
	kl = get_keylist(cfg)

	tx3 = await txsign(cfg,tx2,seed_files,kl,kal)

	if tx3:
		tx3.file.write(ask_write=False)
		tx4 = await SentTX(cfg=cfg, data=tx3.__dict__)
		if await tx4.send():
			tx4.file.write(ask_overwrite=False,ask_write=False)
			tx4.print_contract_addr()
	else:
		die(2,'Transaction could not be signed')

async_run(main())



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# Project source code repository: https://github.com/mmgen/mmgen-wallet
# Licensed according to the terms of GPL Version 3.  See LICENSE for details.

"""
test.cmdtest_py_d.ct_tool: tool tests for the MMGen cmdtest.py test suite
"""

import sys,os

from mmgen.util import suf
from mmgen.color import cyan

from ..include.common import (
	cfg,
	vmsg,
	read_from_file,
	write_to_file,
	cmp_or_die,
	joinpath,
	getrand
)
from .common import hincog_fn,incog_id_fn,hincog_offset,tool_enc_passwd,ref_dir
from .ct_base import CmdTestBase
from .ct_main import CmdTestMain

class CmdTestTool(CmdTestMain,CmdTestBase):
	"interactive 'mmgen-tool' commands"
	networks = ('btc',)
	segwit_opts_ok = False
	tmpdir_nums = [9]
	enc_infn = 'tool_encrypt.in'
	cmd_group = (
		('tool_find_incog_data', (9,"'mmgen-tool find_incog_data'", [[[hincog_fn],1],[[incog_id_fn],1]])),
		('tool_rand2file',       (9,"'mmgen-tool rand2file'", [])),
		('tool_encrypt',         (9,"'mmgen-tool encrypt' (random data)",     [])),
		('tool_decrypt',         (9,"'mmgen-tool decrypt' (random data)", [[[enc_infn+'.mmenc'],9]])),
		('tool_twview_bad_comment',(9,"'mmgen-tool twview' (with bad comment)", [])),
		('tool_decrypt_keystore',(9,"'mmgen-tool decrypt_keystore'", [])),
		('tool_decrypt_geth_keystore',(9,"'mmgen-tool decrypt_geth_keystore'", [])),
		('tool_api',             (9,'tool API (initialization, config methods, wif2addr)',[])),
		# ('tool_encrypt_ref', (9,"'mmgen-tool encrypt' (reference text)",  [])),
	)

	def tool_rand2file(self):
		from mmgen.util2 import parse_bytespec
		for nbytes in ('1','1023','1K','1048575','1M','1048577','123M'):
			t = self.spawn(
				'mmgen-tool',
				['-d',self.tmpdir,'-r0','rand2file','rand2file.out',nbytes],
				extra_desc='({} byte{})'.format( nbytes, suf(parse_bytespec(nbytes)) )
			)
			t.expect('random data written to file')
			t.read()
			t.p.wait()
			t.ok()
		t.skip_ok = True
		return t

	def tool_encrypt(self):
		infile = joinpath(self.tmpdir,self.enc_infn)
		write_to_file(infile,getrand(1033),binary=True)
		t = self.spawn('mmgen-tool',['-d',self.tmpdir,self.usr_rand_arg,'encrypt',infile])
		t.usr_rand(self.usr_rand_chars)
		t.hash_preset('data','1')
		t.passphrase_new('data',tool_enc_passwd)
		t.written_to_file('Encrypted data')
		return t

	def tool_decrypt(self,f1):
		out_fn = 'tool_encrypt.out'
		t = self.spawn('mmgen-tool',['-d',self.tmpdir,'decrypt',f1,'outfile='+out_fn,'hash_preset=1'])
		t.passphrase('data',tool_enc_passwd)
		t.written_to_file('Decrypted data')
		d1 = self.read_from_tmpfile(self.enc_infn,binary=True)
		d2 = self.read_from_tmpfile(out_fn,binary=True)
		cmp_or_die(d1,d2)
		return t

	def tool_find_incog_data(self,f1,f2):
		i_id = read_from_file(f2).rstrip()
		vmsg(f'Incog ID: {cyan(i_id)}')
		t = self.spawn('mmgen-tool',['-d',self.tmpdir,'find_incog_data',f1,i_id])
		o = t.expect_getend(f'Incog data for ID {i_id} found at offset ')
		if not sys.platform == 'win32':
			os.unlink(f1) # causes problems with MSYS2
		cmp_or_die(hincog_offset,int(o))
		return t

	def tool_twview_bad_comment(self): # test correct operation of get_tw_label()
		t = self.spawn(
			'mmgen-tool',
			['twview'],
			env = { 'MMGEN_BOGUS_UNSPENT_DATA': joinpath(ref_dir,'bad-comment-unspent.json') },
			exit_val = 2)
		t.expect('cannot be converted to TwComment')
		return t

	def _decrypt_keystore(self,cmd,fn,pw,chk):
		if cfg.no_altcoin:
			return 'skip'
		t = self.spawn('mmgen-tool',['-d',self.tmpdir,cmd,fn])
		t.expect('Enter passphrase: ',pw+'\n')
		t.expect(chk)
		return t

	def tool_decrypt_keystore(self):
		return self._decrypt_keystore(
			cmd = 'decrypt_keystore',
			fn  = 'test/ref/altcoin/98831F3A-keystore-wallet.json',
			pw = 'abc',
			chk = read_from_file('test/ref/98831F3A.bip39') )

	def tool_decrypt_geth_keystore(self):
		return self._decrypt_keystore(
			cmd = 'decrypt_geth_keystore',
			fn  = 'test/ref/ethereum/geth-wallet.json',
			pw  = '',
			chk = '9627ddb68354f5e0ff45fb2da49d7a20a013b7257a83ef4adbbbd87aeaccc75e')

	def tool_api(self):
		t = self.spawn(
				'tool_api_test.py',
				(['no_altcoin'] if cfg.no_altcoin else []),
				cmd_dir = 'test/misc' )
		t.expect('legacy.*compressed.*segwit.*bech32',regex=True)
		return t



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# Project source code repository: https://github.com/mmgen/mmgen-wallet
# Licensed according to the terms of GPL Version 3.  See LICENSE for details.

"""
test.cmdtest_py_d.ct_input: user input tests for the MMGen cmdtest.py test suite
"""

import sys,os

from mmgen.cfg import gc
from mmgen.util import fmt,capfirst,remove_whitespace
from mmgen.wallet import get_wallet_cls

from ..include.common import (
	cfg,
	imsg,
	imsg_r,
	sample_mn,
	get_data_from_file,
	read_from_file,
	strip_ansi_escapes
)
from .common import Ctrl_U,ref_dir
from .ct_base import CmdTestBase
from .input import stealth_mnemonic_entry,user_dieroll_entry

class CmdTestInput(CmdTestBase):
	'user input'
	networks = ('btc',)
	tmpdir_nums = [1]
	color = True
	cmd_group_in = (
		('subgroup.char', []),
		('subgroup.line', []),
		('subgroup.password', []),
		('subgroup.misc', []),
		('subgroup.wallet', []),
		('subgroup.mnemonic', []),
		('subgroup.dieroll', []),
	)
	cmd_subgroups = {
	'char': (
		'get_char() function',
		('get_char1',                    'get_char()'),
		('get_char2',                    'get_char() [multiple characters]'),
		('get_char3',                    'get_char() [no prompt]'),
		('get_char4',                    'get_char() [utf8]'),
		('get_char_term1',               'get_char() [term, utf8]'),
		('get_char_term2',               'get_char() [term, multiple characters]'),
		('get_char_term3',               'get_char() [term, prehold_protect=False]'),
		('get_char_term4',               'get_char() [term, immed_chars="xyz"]'),
	),
	'line': (
		'line_input() function',
		('line_input',                    'line_input()'),
		('line_input_term1',              'line_input() [term]'),
		('line_input_term2',              'line_input() [term, no hold protect]'),
		('line_input_insert',             'line_input() [inserted text]'),
		('line_input_insert_term1',       'line_input() [inserted text, term]'),
		('line_input_insert_term2',       'line_input() [inserted text, term, no hold protect]'),
		('line_input_edit_term',          'line_input() [inserted + edited text, term, utf8]'),
		('line_input_erase_term',         'line_input() [inserted + erased text, term]'),
	),
	'password': (
		'password entry via line_input()',
		('password_entry_noecho',         'utf8 password entry'),
		('password_entry_noecho_term',    'utf8 password entry [term]'),
		('password_entry_echo',           'utf8 password entry (echoed)'),
		('password_entry_echo_term',      'utf8 password entry (echoed) [term]'),
	),
	'misc': (
		'miscellaneous user-level UI functions',
		('get_seed_from_stdin',           'reading seed phrase from STDIN'),
	),
	'wallet': (
		'hash preset, password and label entry',
		('get_passphrase_ui',             'hash preset, password and label (wallet.py)'),
		('get_passphrase_cmdline',        'hash preset, password and label (wallet.py - from cmdline)'),
		('get_passphrase_crypto',         'hash preset, password and label (crypto.py)'),
	),
	'mnemonic': (
		'mnemonic entry',
		('mnemonic_entry_mmgen',          'stealth mnemonic entry (mmgen)'),
		('mnemonic_entry_mmgen_minimal',  'stealth mnemonic entry (mmgen - minimal entry mode)'),
		('mnemonic_entry_bip39',          'stealth mnemonic entry (bip39)'),
		('mnemonic_entry_bip39_short',    'stealth mnemonic entry (bip39 - short entry mode)'),
		('mn2hex_interactive_mmgen',      'mn2hex_interactive (mmgen)'),
		('mn2hex_interactive_mmgen_fixed','mn2hex_interactive (mmgen - fixed (10-letter) entry mode)'),
		('mn2hex_interactive_bip39',      'mn2hex_interactive (bip39)'),
		('mn2hex_interactive_bip39_short','mn2hex_interactive (bip39 - short entry mode (+pad entry))'),
		('mn2hex_interactive_bip39_fixed','mn2hex_interactive (bip39 - fixed (4-letter) entry mode)'),
		('mn2hex_interactive_xmr',        'mn2hex_interactive (xmrseed)'),
		('mn2hex_interactive_xmr_short',  'mn2hex_interactive (xmrseed - short entry mode)'),
	),
	'dieroll': (
		'dieroll entry',
		('dieroll_entry',                 'dieroll entry (base6d)'),
		('dieroll_entry_usrrand',         'dieroll entry (base6d) with added user entropy'),
	)
	}

	def get_seed_from_stdin(self):
		self.spawn('',msg_only=True)
		from subprocess import run,PIPE
		cmd = ['python3','cmds/mmgen-walletconv','--in-fmt=words','--out-fmt=words','--outdir=test/trash']
		mn = sample_mn['mmgen']['mn']
		run_env = dict(os.environ)
		run_env['MMGEN_TEST_SUITE'] = ''

		# the test can fail the first time if cfg file has changed, so run it twice if necessary:
		for _ in range(2):
			cp = run( cmd, input=mn.encode(), stdout=PIPE, stderr=PIPE, env=run_env )
			if b'written to file' in cp.stderr:
				break

		from mmgen.color import set_vt100
		set_vt100()
		imsg(cp.stderr.decode().strip())
		res = get_data_from_file(cfg,'test/trash/A773B05C[128].mmwords',silent=True).strip()
		assert res == mn, f'{res} != {mn}'
		return 'ok' if b'written to file' in cp.stderr else 'error'

	def get_passphrase_ui(self):
		t = self.spawn('test/misc/get_passphrase.py',['--usr-randchars=0','seed'],cmd_dir='.')

		# 1 - new wallet, default hp,label;empty pw
		t.expect('accept the default.*: ','\n',regex=True)

		# bad repeat
		t.expect('new MMGen wallet: ','pass1\n')
		t.expect('peat passphrase: ','pass2\n')

		# good repeat
		t.expect('new MMGen wallet: ','\n')
		t.expect('peat passphrase: ','\n')
		t.expect('mpty pass')

		t.expect('no label: ','\n')

		t.expect('[][3][No Label]')

		# 2 - new wallet, user-selected hp,pw,label
		t.expect('accept the default.*: ', '1\n', regex=True)

		t.expect('new MMGen wallet: ','pass1\n')
		t.expect('peat passphrase: ','pass1\n')

		t.expect('no label: ','lbl1\n')

		t.expect('[pass1][1][lbl1]')

		# 3 - passchg, nothing changes
		t.expect('new hash preset')
		t.expect('reuse the old value.*: ','\n',regex=True)
		t.expect('unchanged')

		t.expect('new passphrase.*: ','pass1\n',regex=True)
		t.expect('peat new passphrase: ','pass1\n')
		t.expect('unchanged')

		t.expect('reuse the label .*: ','\n',regex=True)
		t.expect('unchanged')

		t.expect('[pass1][1][lbl1]')

		# 4 - passchg, everything changes
		t.expect('new hash preset')
		t.expect('reuse the old value.*: ','2\n',regex=True)
		t.expect(' changed to')

		t.expect('new passphrase.*: ','pass2\n',regex=True)
		t.expect('peat new passphrase: ','pass2\n')
		t.expect(' changed')

		t.expect('reuse the label .*: ','lbl2\n',regex=True)
		t.expect(' changed to')
		t.expect('[pass2][2][lbl2]')

		# 5 - wallet from file
		t.expect('from file')

		# bad passphrase
		t.expect('passphrase for MMGen wallet: ','bad\n')
		t.expect('Trying again')

		# good passphrase
		t.expect('passphrase for MMGen wallet: ','reference password\n')
		t.expect('[reference password][1][No Label]')

		return t

	def get_passphrase_cmdline(self):
		with open('test/trash/pwfile','w') as fp:
			fp.write('reference password\n')
		t = self.spawn('test/misc/get_passphrase.py', [
			'--usr-randchars=0',
			'--label=MyLabel',
			'--passwd-file=test/trash/pwfile',
			'--hash-preset=1',
			'seed' ],
			cmd_dir = '.' )
		for _ in range(4):
			t.expect('[reference password][1][MyLabel]')

		return t

	def get_passphrase_crypto(self):
		t = self.spawn('test/misc/get_passphrase.py',['--usr-randchars=0','crypto'],cmd_dir='.')

		# new passwd
		t.expect('passphrase for .*: ', 'x\n', regex=True)
		t.expect('peat passphrase: ', '\n')
		t.expect('passphrase for .*: ', 'pass1\n', regex=True)
		t.expect('peat passphrase: ', 'pass1\n')
		t.expect('[pass1]')

		# existing passwd
		t.expect('passphrase for .*: ', 'pass2\n', regex=True)
		t.expect('[pass2]')

		# hash preset
		t.expect('accept the default .*: ', '0\n', regex=True)
		t.expect('nvalid')
		t.expect('accept the default .*: ', '8\n', regex=True)
		t.expect('nvalid')
		t.expect('accept the default .*: ', '7\n', regex=True)
		t.expect('[7]')

		# hash preset (default)
		t.expect('accept the default .*: ', '\n', regex=True)
		t.expect(f'[{gc.dfl_hash_preset}]')

		return t

	def _input_func(self,func_name,arg_dfls,func_args,text,expect,term):
		if term and sys.platform == 'win32':
			return ('skip_warn','pexpect_spawn not supported on Windows platform')
		func_args = dict(zip(arg_dfls.keys(),func_args))
		t = self.spawn(
			'test/misc/input_func.py',
			[func_name,repr(func_args)],
			cmd_dir='.',
			pexpect_spawn=term )
		imsg('Parameters:')
		imsg(f'  pexpect_spawn: {term}')
		imsg(f'  sending:       {text!r}')
		imsg(f'  expecting:     {expect!r}')
		imsg('\nFunction args:')
		for k,v in func_args.items():
			imsg('  {:14} {!r}'.format(k+':',v))
		imsg_r('\nScript output: ')
		prompt_add = (func_args['insert_txt'] if term else '') if func_name == 'line_input' else ''
		prompt = func_args['prompt'] + prompt_add
		t.expect('Calling ')
		if prompt:
			t.expect(prompt,text)
		else:
			t.send(text)
		ret = t.expect_getend('  ==> ')
		assert ret == repr(expect), f'Text mismatch! {ret} != {repr(expect)}'
		return t

	def _get_char(self,func_args,text,expect,term):
		arg_dfls = {
			'prompt': '',
			'immed_chars': '',
			'prehold_protect': True,
			'num_bytes': 5,
		}
		return self._input_func('get_char',arg_dfls,func_args,text,expect,term)

	def _line_input(self,func_args,text,expect,term):
		arg_dfls = {
			'prompt': '', # positional
			'echo': True,
			'insert_txt': '',
			'hold_protect': True,
		}
		return self._input_func('line_input',arg_dfls,func_args,text+'\n',expect,term)

	def get_char1(self):
		return self._get_char(['prompt> ','',True,5],'x','x',False)

	def get_char2(self):
		expect = 'x' if sys.platform == 'win32' else 'xxxxx'
		return self._get_char(['prompt> ','',True,5],'xxxxx',expect,False)

	def get_char3(self):
		return self._get_char(['','',True,5],'x','x',False)

	def get_char4(self):
		return self._get_char(['prompt> ','',True,2],'α','α',False) # UTF-8, must get 2 bytes

	def get_char_term1(self):
		return self._get_char(['prompt> ','',True,2],'β','β',True)  # UTF-8, must get 2 bytes

	def get_char_term2(self):
		return self._get_char(['prompt> ','',True,5],'xxxxx','xxxxx',True)

	def get_char_term3(self):
		return self._get_char(['','',False,5],'x','x',True)

	def get_char_term4(self):
		return self._get_char(['prompt> ','xyz',False,5],'x','x',True)

	def line_input(self):
		return self._line_input(['prompt> ',True,'',True],'foo','foo',False)

	def line_input_term1(self):
		return self._line_input(['prompt> ',True,'',True],'foo','foo',True)

	def line_input_term2(self):
		return self._line_input(['prompt> ',True,'',False],'foo','foo',True)

	def line_input_insert(self):
		return self._line_input(['prompt> ',True,'inserted text',True],'foo','foo',False)

	def line_input_insert_term1(self):
		return self._line_input(['prompt> ',True,'foo',True],'bar','foobar',True)

	def line_input_insert_term2(self):
		return self._line_input(['prompt> ',True,'foo',False],'bar','foobar',True)

	def line_input_edit_term(self):
		return self._line_input(['prompt> ',True,'φυφυ',True],'\b\bβαρ','φυβαρ',True)

	def line_input_erase_term(self):
		return self._line_input(['prompt> ',True,'foobarbaz',True],Ctrl_U+'foobar','foobar',True)

	def _password_entry(self,prompt,opts=[],term=False):
		if term and sys.platform == 'win32':
			return ('skip_warn','pexpect_spawn not supported on Windows platform')
		t = self.spawn( 'test/misc/input_func.py', opts + ['passphrase'], cmd_dir='.', pexpect_spawn=term )
		imsg(f'Terminal: {term}')
		pw = 'abc-α'
		t.expect(prompt,pw+'\n')
		ret = t.expect_getend('Entered: ')
		assert ret == pw, f'Password mismatch! {ret} != {pw}'
		return t

	winskip_msg = """
		pexpect_spawn not supported on Windows platform
		Perform the following test by hand with non-ASCII password abc-α
		or another password in your native alphabet:

		  test/misc/input_func.py{} passphrase
	"""

	def password_entry_noecho(self,term=False):
		return self._password_entry('Enter passphrase: ',term=term)

	def password_entry_noecho_term(self):
		if self.skip_for_win():
			return ('skip_warn','\n' + fmt(self.winskip_msg.format(''),strip_char='\t'))
		return self.password_entry_noecho(term=True)

	def password_entry_echo(self,term=False):
		return self._password_entry('Enter passphrase (echoed): ',['--echo-passphrase'],term=term)

	def password_entry_echo_term(self):
		if self.skip_for_win():
			return ('skip_warn','\n' + fmt(self.winskip_msg.format(' --echo-passphrase'),strip_char='\t'))
		return self.password_entry_echo(term=True)

	def _mn2hex(self,fmt,entry_mode='full',mn=None,pad_entry=False,enter_for_dfl=False):
		mn = mn or sample_mn[fmt]['mn'].split()
		t = self.spawn('mmgen-tool',['mn2hex_interactive','fmt='+fmt,'mn_len=12','print_mn=1'])
		from mmgen.mn_entry import mn_entry
		mne = mn_entry( cfg, fmt, entry_mode )
		t.expect(
			'Type a number.*: ',
			('\n' if enter_for_dfl else str(mne.entry_modes.index(entry_mode)+1)),
			regex = True )
		t.expect('Using entry mode (\S+)',regex=True)
		mode = strip_ansi_escapes(t.p.match.group(1)).lower()
		assert mode == mne.em.name.lower(), f'{mode} != {mne.em.name.lower()}'
		stealth_mnemonic_entry(t,mne,mn,entry_mode=entry_mode,pad_entry=pad_entry)
		t.expect(sample_mn[fmt]['hex'])
		return t

	def _user_seed_entry(
			self,
			fmt,
			usr_rand    = False,
			out_fmt     = None,
			entry_mode  = 'full',
			mn          = None,
			seedlen_opt = False):

		wcls = get_wallet_cls(fmt_code=fmt)
		wf = os.path.join(ref_dir,f'FE3C6545.{wcls.ext}')
		if wcls.base_type == 'mnemonic':
			mn = mn or read_from_file(wf).strip().split()
		elif wcls.type == 'dieroll':
			mn = mn or list(remove_whitespace(read_from_file(wf)))
			for idx,val in ((5,'x'),(18,'0'),(30,'7'),(44,'9')):
				mn.insert(idx,val)
		t = self.spawn(
			'mmgen-walletconv',
			['--usr-randchars=10', '--stdout']
			+ (['--seed-len=128'] if seedlen_opt else [])
			+ [f'--in-fmt={fmt}', f'--out-fmt={out_fmt or fmt}']
		)
		t.expect(f'{capfirst(wcls.base_type or wcls.type)} type:.*{wcls.mn_type}',regex=True)
		if not seedlen_opt:
			t.expect(wcls.choose_seedlen_prompt,'1')
			t.expect('(Y/n): ','y')
		if wcls.base_type == 'mnemonic':
			t.expect('Type a number.*: ','6',regex=True)
			t.expect('invalid')
			from mmgen.mn_entry import mn_entry
			mne = mn_entry( cfg, fmt, entry_mode )
			t.expect('Type a number.*: ',str(mne.entry_modes.index(entry_mode)+1),regex=True)
			t.expect('Using entry mode (\S+)',regex=True)
			mode = strip_ansi_escapes(t.p.match.group(1)).lower()
			assert mode == mne.em.name.lower(), f'{mode} != {mne.em.name.lower()}'
			stealth_mnemonic_entry(t,mne,mn,entry_mode=entry_mode)
		elif wcls.type == 'dieroll':
			user_dieroll_entry(t,mn)
			if usr_rand:
				t.expect(wcls.user_entropy_prompt,'y')
				t.usr_rand(10)
			else:
				t.expect(wcls.user_entropy_prompt,'n')
		if not usr_rand:
			sid_chk = 'FE3C6545'
			sid = t.expect_getend(f'Valid {wcls.desc} for Seed ID')
			sid = strip_ansi_escapes(sid.split(',')[0])
			assert sid_chk in sid, f'Seed ID mismatch! {sid_chk} not found in {sid}'
		t.expect('to confirm: ','YES\n')
		return t

	def mnemonic_entry_mmgen_minimal(self):
		from mmgen.mn_entry import mn_entry
		# erase_chars: '\b\x7f'
		m = mn_entry( cfg, 'mmgen', 'minimal' )
		np = 2
		mn = (
			'z',
			'aa',
			'1d2ud',
			'fo{}ot{}#'.format('1' * np, '2' * (m.em.pad_max - np)), # substring of 'football'
			'des1p)%erate\n', # substring of 'desperately'
			'#t!(ie',
			'!)sto8o',
			'the123m8!%s',
			'349t(5)rip',
			'di\b\bdesce',
			'cea',
			'bu\x7f\x7fsuic',
			'app\bpl',
			'wd',
			'busy')
		return self._user_seed_entry('words',entry_mode='minimal',mn=mn)
	def mnemonic_entry_mmgen(self):
		return self._user_seed_entry('words',entry_mode='full')
	def mnemonic_entry_bip39(self):
		return self._user_seed_entry('bip39',entry_mode='full')
	def mnemonic_entry_bip39_short(self):
		return self._user_seed_entry('bip39',entry_mode='short')

	def mn2hex_interactive_mmgen(self):
		return self._mn2hex('mmgen',entry_mode='full')
	def mn2hex_interactive_mmgen_fixed(self):
		return self._mn2hex('mmgen',entry_mode='fixed')
	def mn2hex_interactive_bip39(self):
		return self._mn2hex('bip39',entry_mode='full')
	def mn2hex_interactive_bip39_short(self):
		return self._mn2hex('bip39',entry_mode='short',pad_entry=True)
	def mn2hex_interactive_bip39_fixed(self):
		return self._mn2hex('bip39',entry_mode='fixed',enter_for_dfl=True)
	def mn2hex_interactive_xmr(self):
		return self._mn2hex('xmrseed',entry_mode='full')
	def mn2hex_interactive_xmr_short(self):
		return self._mn2hex('xmrseed',entry_mode='short')

	def dieroll_entry(self):
		return self._user_seed_entry('dieroll', seedlen_opt=True)
	def dieroll_entry_usrrand(self):
		return self._user_seed_entry('dieroll',usr_rand=True,out_fmt='bip39')



#!/usr/bin/env python3

"""
test/unit_tests_d/ut_tx_deserialize: TX deserialization unit tests for the MMGen suite
"""

import os,json

from mmgen.color import purple,cyan
from mmgen.util import msg,Msg,Msg_r
from mmgen.devtools import Pmsg
from mmgen.protocol import init_proto
from mmgen.tx import CompletedTX
from mmgen.proto.btc.tx.base import DeserializeTX
from mmgen.rpc import rpc_init

from ..include.common import cfg,start_test_daemons,stop_test_daemons

def print_info(name,extra_desc):
	if cfg.names:
		Msg_r('{} {} {}'.format(
			purple('Testing'),
			cyan(f'{name} ({extra_desc})'),
			'' if cfg.quiet else '\n'))
	else:
		Msg_r(f'Testing {extra_desc}')
		if not cfg.quiet:
			Msg('')

async def test_tx(tx_proto,tx_hex,desc,n):

	def has_nonstandard_outputs(outputs):
		for o in outputs:
			t = o['scriptPubKey']['type']
			if t in ('nonstandard','pubkey','nulldata'):
				return True
		return False

	rpc = await rpc_init( cfg, proto=tx_proto, ignore_wallet=True )
	d = await rpc.call('decoderawtransaction',tx_hex)

	if has_nonstandard_outputs(d['vout']):
		return False

	dt = DeserializeTX(tx_proto,tx_hex)

	if cfg.verbose:
		Msg('\n\n================================ Core vector: ==================================')
	Msg_r('.' if cfg.quiet else f'{n:>3}) {desc}\n')
	if cfg.verbose:
		Pmsg(d)
		Msg('\n------------------------------ MMGen deserialized: -----------------------------')
		Pmsg(dt._asdict())

	# metadata
	assert dt.txid == d['txid'],'TXID does not match'
	assert dt.locktime == d['locktime'],'Locktime does not match'
	assert dt.version == d['version'],'Version does not match'

	# inputs
	a,b = d['vin'],dt.txins
	for i in range(len(a)):
		assert a[i]['txid'] == b[i]['txid'],f'TxID of input {i} does not match'
		assert a[i]['vout'] == b[i]['vout'],f'vout of input {i} does not match'
		assert a[i]['sequence'] == int(b[i]['nSeq'],16),(
			f'nSeq of input {i} does not match')
		if 'txinwitness' in a[i]:
			assert a[i]['txinwitness'] == b[i]['witness'],(
				f'witness of input {i} does not match')

	# outputs
	a,b = d['vout'],dt.txouts
	for i in range(len(a)):
		if 'addresses' in a[i]['scriptPubKey']:
			A = a[i]['scriptPubKey']['addresses'][0]
			B = b[i]['address']
			fs = 'address of output {} does not match\nA: {}\nB: {}'
			assert A == B, fs.format(i,A,B)

		A = a[i]['value']
		B = b[i]['amount']
		fs = 'value of output {} does not match\nA: {}\nB: {}'
		assert A == B, fs.format(i,A,B)

		A = a[i]['scriptPubKey']['hex']
		B = b[i]['scriptPubKey']
		fs = 'scriptPubKey of output {} does not match\nA: {}\nB: {}'
		assert A == B, fs.format(i,A,B)

async def do_mmgen_ref(daemons,fns,name,desc):
	# NB: remove_datadir is required here for some reason (seems to be Bitcoin Core version-dependent)
	start_test_daemons(*daemons,remove_datadir=True)
	print_info(name,desc)
	for n,fn in enumerate(fns):
		tx = await CompletedTX( cfg=cfg, filename=fn, quiet_open=True )
		await test_tx(
			tx_proto = tx.proto,
			tx_hex   = tx.serialized,
			desc     = fn,
			n        = n+1 )
	stop_test_daemons(*daemons)
	Msg('OK')
	return True

class unit_tests:

	altcoin_deps = ('mmgen_ref_alt',)

	async def core_vectors(self,name,ut):

		core_repo_root = os.getenv('CORE_REPO_ROOT')
		if not core_repo_root:
			msg('The environmental variable CORE_REPO_ROOT must be set before running this test')
			return False

		start_test_daemons('btc')

		fn_b = 'src/test/data/tx_valid.json'
		fn = os.path.join(core_repo_root,fn_b)
		with open(fn) as fp:
			core_data = json.loads(fp.read())
		print_info(name,'Bitcoin Core test vectors')
		n = 1
		for e in core_data:
			if isinstance(e[0],list):
				await test_tx(
					tx_proto = init_proto( cfg, 'btc', need_amt=True ),
					tx_hex   = e[1],
					desc     = desc,
					n        = n )
				n += 1
			else:
				desc = e[0]

		Msg('OK')
		stop_test_daemons('btc')
		return True

	async def mmgen_ref(self,name,ut):
		return await do_mmgen_ref(
			('btc','btc_tn'),
			(
				'test/ref/0B8D5A[15.31789,14,tl=1320969600].rawtx',
				'test/ref/0C7115[15.86255,14,tl=1320969600].testnet.rawtx',
				'test/ref/542169[5.68152,34].sigtx',
			),
			name,
			'MMGen reference transactions [Bitcoin]' )

	async def mmgen_ref_alt(self,name,ut):
		return await do_mmgen_ref(
			('ltc','ltc_tn','bch'),
			(
				'test/ref/litecoin/AF3CDF-LTC[620.76194,1453,tl=1320969600].rawtx',
				'test/ref/litecoin/A5A1E0-LTC[1454.64322,1453,tl=1320969600].testnet.rawtx',
				'test/ref/460D4D-BCH[10.19764,tl=1320969600].rawtx'
			),
			name,
			'MMGen reference transactions [Altcoin]' )



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_py_d.ot_btc_testnet: BTC testnet test vectors for MMGen data objects
"""

from mmgen.key import PrivKey,WifKey
from mmgen.addr import CoinAddr
from mmgen.protocol import init_proto

from .ot_common import r16,r32
from ..include.common import cfg

proto = init_proto( cfg, 'btc', network='testnet', need_amt=True )

tests = {
	'CoinAddr': {
		'bad':  (
			{'addr':1,   'proto':proto},
			{'addr':'x', 'proto':proto},
			{'addr':'я', 'proto':proto},
		),
		'good':  (
			{'addr':'n2FgXPKwuFkCXF946EnoxWJDWF2VwQ6q8J', 'proto':proto},
			{'addr':'2MspvWFjBbkv2wzQGqhxJUYPCk3Y2jMaxLN','proto':proto},
		),
	},
	'WifKey': {
		'bad': (
			{'proto':proto, 'wif':1},
			{'proto':proto, 'wif':[]},
			{'proto':proto, 'wif':'\0'},
			{'proto':proto, 'wif':'\1'},
			{'proto':proto, 'wif':'я'},
			{'proto':proto, 'wif':'g'},
			{'proto':proto, 'wif':'gg'},
			{'proto':proto, 'wif':'FF'},
			{'proto':proto, 'wif':'f00'},
			{'proto':proto, 'wif':r16.hex()},
			{'proto':proto, 'wif':'2MspvWFjBbkv2wzQGqhxJUYPCk3Y2jMaxLN'},
		),
		'good': (
			{'proto':proto, 'wif':'93HsQEpH75ibaUJYi3QwwiQxnkW4dUuYFPXZxcbcKds7XrqHkY6',  'ret_idx':1},
			{'proto':proto, 'wif':'cMsqcmDYZP1LdKgqRh9L4ZRU9br28yvdmTPwW2YQwVSN9aQiMAoR', 'ret_idx':1},
		)
	},
	'PrivKey': {
		'bad': (
			{'proto':proto, 'wif':1},
			{'proto':proto, 'wif':'1'},
			{'proto':proto, 'wif':'5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb'},
			{'proto':proto, 's':r32,'wif':'5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb'},
			{'proto':proto, 'pubkey_type':'std'},
			{'proto':proto, 's':r32},
			{'proto':proto, 's':r32,'compressed':'yes'},
			{'proto':proto, 's':r32,'compressed':'yes','pubkey_type':'std'},
			{'proto':proto, 's':r32,'compressed':True,'pubkey_type':'nonstd'},
			{'proto':proto, 's':r32+b'x','compressed':True,'pubkey_type':'std'}
		),
		'good': (
			{'proto':proto, 'wif':'93HsQEpH75ibaUJYi3QwwiQxnkW4dUuYFPXZxcbcKds7XrqHkY6',
			'ret':bytes.fromhex('e0aef965b905a2fedf907151df8e0a6bac832aa697801c51f58bd2ecb4fd381c')},
			{'proto':proto, 'wif':'cMsqcmDYZP1LdKgqRh9L4ZRU9br28yvdmTPwW2YQwVSN9aQiMAoR',
			'ret':bytes.fromhex('08d0ed83b64b68d56fa064be48e2385060ed205be2b1e63cd56d218038c3a05f')},
			{'proto':proto, 's':r32,'compressed':False,'pubkey_type':'std','ret':r32},
			{'proto':proto, 's':r32,'compressed':True,'pubkey_type':'std','ret':r32}
		),
	},
}



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.wallet: Wallet routines for the 'mmgen-tool' utility
"""

from .common import tool_cmd_base

from ..subseed import SubSeedList
from ..seedsplit import MasterShareIdx
from ..wallet import Wallet

class tool_cmd(tool_cmd_base):
	"key, address or subseed generation from an MMGen wallet"

	def __init__(self,cfg,cmdname=None,proto=None,mmtype=None):
		self.need_proto = cmdname in ('gen_key','gen_addr')
		super().__init__(cfg,cmdname=cmdname,proto=proto,mmtype=mmtype)

	def _get_seed_file(self,wallet):
		from ..fileutil import get_seed_file
		return get_seed_file(
			cfg     = self.cfg,
			wallets = [wallet] if wallet else [],
			nargs   = 1 )

	def get_subseed(self,subseed_idx:str,wallet=''):
		"get the Seed ID of a single subseed by Subseed Index for default or specified wallet"
		self.cfg.quiet = True
		return Wallet(self.cfg,self._get_seed_file(wallet)).seed.subseed(subseed_idx).sid

	def get_subseed_by_seed_id(self,seed_id:str,wallet='',last_idx=SubSeedList.dfl_len):
		"get the Subseed Index of a single subseed by Seed ID for default or specified wallet"
		self.cfg.quiet = True
		ret = Wallet(self.cfg,self._get_seed_file(wallet)).seed.subseed_by_seed_id( seed_id, last_idx )
		return ret.ss_idx if ret else None

	def list_subseeds(self,subseed_idx_range:str,wallet=''):
		"list a range of subseed Seed IDs for default or specified wallet"
		self.cfg.quiet = True
		from ..subseed import SubSeedIdxRange
		return Wallet(self.cfg,self._get_seed_file(wallet)).seed.subseeds.format(
			*SubSeedIdxRange(subseed_idx_range) )

	def list_shares(self,
			share_count: int,
			id_str = 'default',
			master_share: f'(min:1, max:{MasterShareIdx.max_val}, 0=no master share)' = 0,
			wallet = '' ):
		"list the Seed IDs of the shares resulting from a split of default or specified wallet"
		self.cfg.quiet = True
		return Wallet(self.cfg,self._get_seed_file(wallet)).seed.split(
			share_count, id_str, master_share ).format()

	def gen_key(self,mmgen_addr:str,wallet=''):
		"generate a single WIF key for specified MMGen address from default or specified wallet"
		return self._gen_keyaddr( mmgen_addr, 'wif', wallet )

	def gen_addr(self,mmgen_addr:str,wallet=''):
		"generate a single MMGen address from default or specified wallet"
		return self._gen_keyaddr( mmgen_addr, 'addr', wallet )

	def _gen_keyaddr(self,mmgen_addr,target,wallet=''):
		from ..addr import MMGenID
		from ..addrlist import AddrList,AddrIdxList

		addr = MMGenID( self.proto, mmgen_addr )
		self.cfg.quiet = True
		ss = Wallet(self.cfg,self._get_seed_file(wallet))

		if ss.seed.sid != addr.sid:
			from ..util import die
			die(1,f'Seed ID of requested address ({addr.sid}) does not match wallet ({ss.seed.sid})')

		d = AddrList(
			cfg       = self.cfg,
			proto     = self.proto,
			seed      = ss.seed,
			addr_idxs = AddrIdxList(str(addr.idx)),
			mmtype    = addr.mmtype,
			skip_chksum = True ).data[0]

		return { 'wif': d.sec.wif, 'addr': d.addr }[target]



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, a command-line cryptocurrency wallet
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.etc.params: Ethereum Classic protocol
"""

from ..eth.params import mainnet

class mainnet(mainnet):
	chain_names = ['classic','ethereum_classic']
	max_tx_fee  = '0.005'
	coin_amt    = 'ETCAmt'
	ignore_daemon_version = False

class testnet(mainnet):
	chain_names = ['morden','morden_testnet','classic-testnet']

class regtest(testnet):
	chain_names = ['developmentchain']
